# -*- coding: utf-8 -*-

"""
stitchrfunctions.py

Functions for stiTChR and its related scripts
"""

import argparse
import collections as coll
import gzip
import os
import re
import sys
import textwrap
import datetime
import warnings

# Ensure correct importlib-resources function imported
if sys.version_info < (3, 9):
    import importlib_resources                              # PyPI
else:
    import importlib.resources as importlib_resources       # importlib.resources

__version__ = '1.4.0'
__author__ = 'Jamie Heather'
__email__ = 'jheather@mgh.harvard.edu'

# sys.tracebacklimit = 0  # comment when debugging  # TODO fix

data_files = importlib_resources.files("Data")
additional_genes_file = str(data_files / 'additional-genes.fasta')
linkers_file = str(data_files / 'linkers.tsv')
data_dir = os.path.dirname(additional_genes_file)
gui_examples_dir = os.path.join(data_dir, 'GUI-Examples')


def custom_formatwarning(warning_msg, *args, **kwargs):
    """
    Function to make warnings.warn output just the warning text, not the underlying code
    See https://stackoverflow.com/questions/2187269/print-only-the-message-on-warnings
    """
    return str(warning_msg) + '\n'


def read_fa(ff):
    """
    :param ff: opened fasta file
    read_fa(file):Heng Li's Python implementation of his readfq function (tweaked to only bother with fasta)
    https://github.com/lh3/readfq/blob/master/readfq.py
    """

    last = None                                 # this is a buffer keeping the last unprocessed line
    while True:                                 # mimic closure
        if not last:                            # the first record or a record following a fastq
            for l in ff:                        # search for the start of the next record
                if l[0] in '>':                 # fasta header line
                    last = l[:-1]               # save this line
                    break
        if not last:
            break
        name, seqs, last = last[1:], [], None
        for l in ff:                            # read the sequence
            if l[0] in '>':
                last = l[:-1]
                break
            seqs.append(l[:-1])
        if not last or last[0] != '+':          # this is a fasta record
            yield name, ''.join(seqs), None     # yield a fasta record
            if not last:
                break
        else:
            raise IOError("Input file does not appear to be a FASTA file - please check and try again")


def fasta_to_dict(path_to_file):
    """
    :param path_to_file: str of path to FASTA file
    :return: a coll.defaultdict in the format {fasta_header: fasta_read, ...}
    """

    if not os.path.isfile(path_to_file):
        raise IOError("Cannot locate provided FASTA file, " + path_to_file)

    out_dict = coll.defaultdict()
    with open(path_to_file, 'r') as in_file:
        for header, seq, null in read_fa(in_file):
            out_dict[header] = seq

    return out_dict


def fastafy(gene, seq_line):
    """
    :param gene: Gene symbol, extracted from the read id
    :param seq_line: Total protein primary sequence, extracted from input FASTA/generated by in silico splicing
    :return: An output-compatible FASTA entry ready for writing to file
    """
    return ">" + gene + "\n" + textwrap.fill(seq_line, 60) + "\n"


def today(date_format=''):
    """
    :return: Today's date, in one of two formats, defaulting to ISO
    """
    date_today = datetime.datetime.today().date().isoformat()
    if date_format == 'ncbi':
        date_today = datetime.datetime.strptime(date_today, '%Y-%m-%d').strftime('%d-%b-%Y').upper()
    return date_today


def nest():
    """
    Create nested defaultdicts
    """
    return coll.defaultdict(list)


def nest_counter():
    """
    Create nested counters
    """
    return coll.Counter()


def get_chain(v, j):
    """
    :param v: str describing V gene (from input args)
    :param j: str describing J gene (from input args)
    :return: str of the chain in use (if plausible) or throw an error
    """

    if v.startswith('TRB') and j.startswith('TRB'):
        return 'TRB'
    elif (v.startswith('TRA') or v.startswith('TRD')) and j.startswith('TRA'):
        return 'TRA'
    elif v.startswith('TRG') and j.startswith('TRG'):
        return 'TRG'
    elif v.startswith('TRD') and j.startswith('TRD'):
        return 'TRD'

    elif v.startswith('IGH') and j.startswith('IGH'):
        return 'IGH'
    elif v.startswith('IGL') and j.startswith('IGL'):
        return 'IGL'
    elif v.startswith('IGK') and j.startswith('IGK'):
        return 'IGK'

    else:
        raise ValueError("Please ensure you're providing full IMGT gene names (allele number optional), " +
                         "both from the same chain (alpha or beta). Should start \'TR[ABGD]\'.")


def find_species_covered():
    """
    :return: list of data directories for different species available
    """
    species_list = [x for x in os.listdir(data_dir) if os.path.isdir(os.path.normpath(os.path.join(data_dir, x))) and
                    x != 'kazusa' and x != 'GUI-Examples' and '__' not in x]

    if not species_list:
        raise ValueError("No species data detected. Please run stitchrdl first, or otherwise install data in the Data"
                         " directory (" + data_dir + ").")
    species_list.sort()
    return species_list


def infer_species(path_to_file):
    """
    :param path_to_file: str of a path to an input file that may or may not contain a species str
    :return: the species detected, if one found that fits the data available in the data directory
    """
    in_file_name = os.path.basename(path_to_file)
    species_search = [x for x in find_species_covered() if x in in_file_name.upper()]

    if len(species_search) == 1:
        return species_search[0]
    else:
        return ''


def sort_input(cmd_line_args):
    """
    :param cmd_line_args: vars(args())
    :return: tidied/autofilled input arguments, plus the chain information
    """
    tidied_args = tidy_input(cmd_line_args)
    species_dirs = find_species_covered()

    # Check the species is an appropriate one
    if tidied_args['species'] not in species_dirs:
        raise ValueError("Invalid species option given. Current options in Data directory are: "
                         + ', '.join(species_dirs))

    # If additional optional 5'/3' sequences are provided, check they are valid DNA sequences
    for end in ['5', '3']:
        if cmd_line_args[end + '_prime_seq']:
            if not dna_check(cmd_line_args[end + '_prime_seq']):
                raise IOError("Provided " + end + "\' sequence contains non-DNA characters.")

            if len(cmd_line_args[end + '_prime_seq']) % 3 != 0:
                warnings.warn(
                    "Warning: length of " + end + "\' sequence provided is not divisible by 3. "
                                                  "Ensure sequence is padded properly if needed to be in frame.")

    chain = get_chain(tidied_args['v'], tidied_args['j'])
    finished_args = autofill_input(tidied_args, chain)
    return finished_args, chain


def tidy_input(cmd_line_args):
    """
    :param cmd_line_args: vars(args())
    :return: input arguments, with all keys made uppercase (unless it's a path)
    """

    out_args = {}
    for arg in cmd_line_args:

        if isinstance(cmd_line_args[arg], str):
            if 'path' in arg:
                out_args[arg] = cmd_line_args[arg]
            else:
                out_args[arg] = cmd_line_args[arg].upper().replace(' ', '')
        else:
            out_args[arg] = cmd_line_args[arg]

    if cmd_line_args['name']:
        out_args['name'] = cmd_line_args['name']
    else:
        out_args['name'] = ''

    return out_args


def autofill_input(cmd_line_args, chain):
    """
    :param cmd_line_args: Tidied input arguments
    :param chain: TCR chain, just determined
    :return: Autofilled input arguments (i.e. filling out the leader and constant region genes)
    """

    # Default constant regions for humans and mice
    if not cmd_line_args['c']:
        if cmd_line_args['species'] in ['HUMAN', 'MOUSE']:
            if chain == 'TRA':
                cmd_line_args['c'] = 'TRAC*01'
            elif chain == 'TRB':
                if 'TRBJ1' in cmd_line_args['j']:
                    cmd_line_args['c'] = 'TRBC1*01'
                elif 'TRBJ2' in cmd_line_args['j']:
                    cmd_line_args['c'] = 'TRBC2*01'
            elif chain == 'TRD':
                cmd_line_args['c'] = 'TRDC*01'
            elif chain == 'TRG':
                if cmd_line_args['species'] == 'HUMAN':
                    if 'TRGJ2' in cmd_line_args['j'] or 'TRGJP2' in cmd_line_args['j']:
                        cmd_line_args['c'] = 'TRGC2*02'
                    else:
                        cmd_line_args['c'] = 'TRGC1*01'
                elif cmd_line_args['species'] == 'MOUSE':
                    if 'TRGJ1' in cmd_line_args['j']:
                        cmd_line_args['c'] = 'TRGC1*01'
                    elif 'TRGJ2' in cmd_line_args['j']:
                        cmd_line_args['c'] = 'TRGC2*cmd_line01'
                    elif 'TRGJ3' in cmd_line_args['j']:
                        cmd_line_args['c'] = 'TRGC3*01'
                    elif 'TRGJ4' in cmd_line_args['j']:
                        cmd_line_args['c'] = 'TRGC4*01'

        else:
            raise IOError("Constant region cannot be automatically inferred for non-human/non-mouse TCRs - "
                          "please explicitly specify the relevant constant region.")

    # Leader region (assuming the proximal L for that V is used)
    if not cmd_line_args['l']:
        cmd_line_args['l'] = cmd_line_args['v']

    return cmd_line_args


def get_ref_data(tcr_chain, gene_types, species):
    """
    :param tcr_chain: 3 digit str code, e.g. TRA or TRB
    :param gene_types: list of TYPES of genes to be expected in a final TCR mRNA, in their IMGT nomenclature
    :param species: upper case str, for use if a specific absolute path not specified
    :return: triply nested dict of TCR data: { region { gene { allele { seq } } }; a doubly nested dict with V/J
      functionalities, and a doubly nested dict of genes filtered out due to being partial in their 5' or 3' (or both)
    """

    # Run some basic sanity/input file checks
    if tcr_chain not in ['TRA', 'TRB', 'TRG', 'TRD', 'IGH', 'IGL', 'IGK']:
        raise ValueError("Incorrect chain detected (" + tcr_chain + "), cannot get reference data. ")

    in_file_path = os.path.join(data_dir, species, tcr_chain + '.fasta')
    if not os.path.isfile(in_file_path):
        raise IOError(tcr_chain + '.fasta not detected in the Data directory. '
                                  'Please check data exists for this species/locus combination. ')

    # Read in the data to a nested dict
    tcr_data = {}
    for gene_type in gene_types:
        tcr_data[gene_type] = coll.defaultdict(nest)

    functionality = coll.defaultdict(nest)
    partial_genes = coll.defaultdict(nest)

    with open(in_file_path, 'r') as in_file:
        for fasta_id, seq, blank in read_fa(in_file):
            bits = fasta_id.split('|')
            if len(bits) < 13:
                raise IOError("Input TCR FASTA file does not fit the IMGT header format. ")

            gene, allele = bits[1].split('*')
            functionality_call = bits[3]
            seq_type = fasta_id.split('~')[1]
            partial_flag = bits[13]

            functionality[gene][allele] = functionality_call

            if 'partial' in partial_flag:
                partial_genes[gene][allele] = partial_flag
            else:
                tcr_data[seq_type][gene][allele] = seq.upper()

    for gene_type in gene_types:
        if len(tcr_data[gene_type]) == 0:
            raise Exception("No entries for " + gene_type + " in IMGT data. ")

    return tcr_data, functionality, partial_genes


def strip_functionality(functionality_str):
    """
    :param functionality_str: functionality string as present in an IMGT FASTA header
    :return: the core functionality (F/ORF/P) minus any brackets
    """
    return functionality_str.replace('(', '').replace(')', '').replace('[', '').replace(']', '')


def get_additional_genes(imgt_data, imgt_functionality):
    """
    :param imgt_data: the nested dict produced by get_ref_data containing V/J/C sequence data
    :param imgt_functionality: the nested dict with imgt_stated functionality
    :return: the same dicts supplemented with any genes found in the 'additional genes.fasta' file
    """

    with open(additional_genes_file, 'r') as in_file:
        for fasta_id, seq, blank in read_fa(in_file):
            bits = fasta_id.split('|')

            if len(bits) < 5:
                raise IOError("Sequence in additional-genes.fasta doesn't have enough fields in header: " + fasta_id)

            if '*' in bits[1]:
                gene, allele = bits[1].upper().split('*')
            else:
                raise IOError("Sequence in additional-genes.fasta doesn't have correct gene name format ('" + bits[1]
                              + "'): " + fasta_id)

            if bits[3]:
                functionality_call = bits[3].replace('(', '').replace(')', '').replace('[', '').replace(']', '')
            else:
                functionality_call = 'F'

            if '~' in fasta_id:
                seq_type = fasta_id.split('~')[1]
                if seq_type not in regions.values():
                    raise IOError("Sequence in additional-genes.fasta doesn't have valid gene type ('" + seq_type + "')"
                                  ": " + fasta_id)
            else:
                raise IOError("Sequence in additional-genes.fasta doesn't have the required '~' character: " + fasta_id)

            imgt_data[seq_type][gene][allele] = seq
            imgt_functionality[gene][allele] = functionality_call

    return imgt_data, imgt_functionality


def get_preferred_alleles(path_to_pa_file, gene_types, imgt_data, partiality, locus):
    """
    :param path_to_pa_file: str path to file of preferred alleles
    :param gene_types: list of TYPES of genes to be expected in a final TCR mRNA, in their IMGT nomenclature
    :param imgt_data: double nested dict of tcr gene sequences, from get_ref_data()
    :param partiality: double nested dict detailing partial genes, from get_ref_data()
    :param locus: three character string detailing what locus is being stitched
    :return: a nested dict containing default alleles for the specified genes
    """

    if not os.path.isfile(path_to_pa_file):
        raise IOError("Could not find a preferred allele file at this path: " + path_to_pa_file
                      + ", despite the '-p' flag being used. Please check path. ")

    preferences = {}
    for gene_type in gene_types:
        preferences[gene_type] = coll.defaultdict(nest)

    line_count = 0
    with open(path_to_pa_file, 'r') as in_file:
        for line in in_file:
            bits = line.rstrip().split('\t')
            if line_count == 0:
                header = bits
            else:
                pref_gene, pref_allele, pref_region, pref_locus, pref_source = bits

                base_warning = "Requested preferred allele " + pref_gene + "*" + pref_allele + " cannot be used for " \
                               "the " + pref_region + " region, "

                # Only process those which are labelled as being involved in the locus under consideration
                covered_loci = pref_locus.replace(' ', '').split(',')
                if locus not in covered_loci:
                    # Silently ignore other loci's preferred alleles
                    continue

                # Check that provided preference is valid, present in the data, and not a partial gene
                if pref_region not in regions.values():
                    warnings.warn(base_warning +
                                  "as this is not a valid option. ")
                elif pref_gene not in imgt_data[pref_region]:
                    warnings.warn(base_warning +
                                  "as this gene is not present in the input FASTA data for this species. ")
                elif pref_allele not in imgt_data[pref_region][pref_gene]:
                    warnings.warn(base_warning +
                                  "as this allele is not present in the input FASTA data for this species. ")
                elif pref_allele in partiality[pref_gene][pref_allele]:
                    warnings.warn(base_warning +
                                  "as the sequence for this allele is in the input data is flagged as partial. ")
                else:
                    preferences[pref_region][pref_gene] = pref_allele

            line_count += 1

    return preferences


def tidy_n_term(n_term_nt):
    """
    Tidy up the germline N-terminal half (i.e. pre-CDR3, L+V) of the nt seq so that it's nicely divisible by 3
    :param n_term_nt: done['l'] + done['v']
    :return: n_term_nt trimmed to no hanging fragments
    """

    modulo = len(n_term_nt) % 3
    if modulo == 0:
        trimmed = n_term_nt
    else:
        trimmed = n_term_nt[:-modulo]

    return trimmed, translate_nt(trimmed)


def find_stop(seq):
    """
    :param seq: a translated amino acid sequence
    :return: the position of the first stop codon (*) inside - if none, return length of whole sequence
    """
    if '*' in seq:
        return seq.index('*')
    else:
        return len(seq)


def tidy_c_term(c_term_nt, skip, c_region_motifs, c_gene):
    """
    Tidy up the germline C-terminal half (i.e. post-CDR3, J+C) of the nt seq so that it's the right frame/trimmed
    :param c_term_nt: done['j'] + done['c']
    :param skip: boolean, whether to skip the C region checks or not
    :param c_region_motifs: nested dict of details of motifs required for finding the right translation frame
    :param c_gene: str of actual C gene used
    :return: c_term_nt trimmed/in right frame
    """

    translations = {}  # key = frame (0, 1, or 2)
    positions = {}
    stop_counts = {}
    motif_counts = {}

    if c_gene not in c_region_motifs['start']:
        skip = True

    # Try every frame, look for the frame that has the right properties (motif, lack of stops etc)
    for f in range(3):
        # TODO another possible thimble performance boost opportunity: store previous translations for a C within a run?
        translated = translate_nt(c_term_nt[f:])
        translations[f] = translated
        positions[f] = find_stop(translated)
        stop_counts[f] = len([x for x in translated if x == '*'])

        if not skip or c_gene in c_region_motifs['start']:
            # Use the defined constant region motif
            if c_region_motifs['start'][c_gene] in translated:

                # Account for late EX4UTR stop codons
                if c_gene in c_region_motifs['stop']:
                    stop_index_aa = translated.index(c_region_motifs['stop'][c_gene])
                    c_term_nt = c_term_nt[:(stop_index_aa * 3) + f]  # And offset by the frame to prevent trailing nt
                    translated = translate_nt(c_term_nt[f:])

                    if 'UTR' not in c_region_motifs['exons'][c_gene]:
                        warnings.warn("Warning: the constant region '" + c_gene + "' being used contains a stop codon, "
                                      "yet its exon label (" + c_region_motifs['exons'][c_gene] + ") doesn't suggest "
                                      "there should be one - this could indicate incorrect exon annotations, "
                                      "potentially resulting in an out of frame constant region. ")

                return c_term_nt[f:], translated

        else:
            # Just look for ALL species-related C region translated motifs
            motif_counts[f] = len([x for x in c_region_motifs['start'].values() if x in translated])

    # If the code has made it to here, it hasn't made an unambiguous C terminus frame determination, and needs to choose
    frames_with_c_motifs = [x for x in motif_counts if motif_counts[x] > 0]
    stopless = [x for x in stop_counts if stop_counts[x] == 0]
    early_stops = [x for x in range(3) if stop_counts[x] == 1 and positions[x] < 3]

    # First prioritise a single frame that contains a known TCR constant region start motif
    if len(frames_with_c_motifs) == 1:
        f = frames_with_c_motifs[0]
        warnings.warn("Note: a single reading frame for translating the C terminus with any relevant known C region "
                      "motif was detected. ")

    # Failing that, if there's only one frame without stop codons, choose that frame
    elif len(stopless) == 1:
        f = stopless[0]
        warnings.warn("Note: no C region motifs were detected: selected the single reading frame for translating the C "
                      "terminus with no stop codons. ")

    # Some TCR J genes contain a 5' stop that can be removed during recombination, so prioritise those next
    elif len(early_stops) == 1:
        f = early_stops[0]
        warnings.warn("Note: no C region motifs or stop-codon-less translation frames detected: the single frame with "
                      "a 5' recombination-deletable stop codon was detected. ")

    # Finally pick the frame that gives the longest sequence before a stop
    else:
        f = max(positions, key=positions.get)
        warnings.warn("Note: no obvious C region frame detected: choosing the frame giving the longest ORF. ")

    return c_term_nt[f:], translations[f]


def determine_v_interface(cdr3aa, n_term_nuc, n_term_amino):
    """
    Determine germline V contribution, and subtract from the CDR3 (to leave just non-templated residues)
    :param cdr3aa: CDR3 region (protein sequence as provided)
    :param n_term_nuc: DNA encoding the germline N terminal portion (i.e. L1+2 + V gene), with no untranslated bp
    :param n_term_amino: translation of n_term_nuc
    :return: appropriately trimmed n_term_nuc, plus the number of AA residues the CDR3's N term can be trimmed by
    """

    for c in reversed(list(range(1, 5))):
        n_term_cdr3_chunk = cdr3aa[:c]
        for v in range(10):
            aa_l = len(n_term_amino)
            v_match = n_term_amino[aa_l - (c + v):aa_l - v]
            if n_term_cdr3_chunk == v_match:
                n_term_nt_trimmed = n_term_nuc[:(aa_l * 3) - (v * 3)]
                cdr3_n_offset = c
                return n_term_nt_trimmed, cdr3_n_offset

    # Shouldn't be able to throw an error, as the presence of an N terminal cysteine should be established, but in case
    raise Exception("Unable to locate N terminus of CDR3 in V gene correctly. Please ensure sequence plausibility. ")


def find_cdr3_c_term(cdr3_chunk, j_seq, strict):
    """
    Take a C terminal section of a CDR3 and find its germline contribution from the J (using the  pattern)
    :param cdr3_chunk: a portion of the CDR3 junction, having had V germline contributions removed
    :param j_seq: the sequence of the germline J gene in which should lie the cdr3_chunk (or a subsequent substring)
    :param strict: boolean flag determining whether strict (GXG) or non-strict (G..|..G) pattern is used
    :yield: the index of all hits of cdr3_chunk followed by G or XXG in j_seq (i.e. nesting it at the conserved spot)
    """
    i = j_seq.find(cdr3_chunk)
    while i != -1:

        if strict:
            if j_seq[i + len(cdr3_chunk)] == 'G' and j_seq[i + len(cdr3_chunk) + 2] == 'G':
                yield i

        else:
            if j_seq[i + len(cdr3_chunk)] == 'G':
                yield i
            elif j_seq[i + len(cdr3_chunk) + 2] == 'G':
                yield i

        i = j_seq.find(cdr3_chunk, i+1)


def determine_j_interface(cdr3_cterm_aa, c_term_nuc, c_term_amino, gl_nt_j_len, j_warning_threshold):
    """
    Determine germline J contribution, and subtract from the CDR3 (to leave just non-templated residues)
    Starts with the whole CDR3 (that isn't contributed by V) and looks for successively N-terminal truncated
    regions in the germline J-REGION.
    :param cdr3_cterm_aa: CDR3 region (protein sequence), minus the N-terminal V gene germline contributions
    :param c_term_nuc: DNA encoding the germline C terminal portion (i.e. J + C genes), with no untranslated bp
    :param c_term_amino: translation of c_term_nuc (everything downstream of recognisable end of the V)
    :param gl_nt_j_len: length of germline nucleotide J gene used (to only search for CDR3 C terminal in J)
    :param j_warning_threshold: int threshold value, if a J substring length match is shorter it will throw a warning
    :return: the nt seq of the C-terminal section of the TCR and number of bases into CDR3 that are non-templated
    """

    # Figure out how long into the C terminus we need to look, i.e. as far as the J gene stretches
    search_len = int(gl_nt_j_len / 3)

    # Determine germline J contribution - going for longest possible, starting with whole CDR3
    # 'c' here is effectively the number of CDR3 AA not contributed to from J (and V, from prior steps)
    for c in reversed(list(range(1, len(cdr3_cterm_aa)))):
        c_term_cdr3_chunk = cdr3_cterm_aa[-c:]

        # Look for the decreasing chunks of the CDR3 in the theoretical translation of this J gene as germline
        search = [x for x in find_cdr3_c_term(c_term_cdr3_chunk, c_term_amino[:search_len], False)]
        if search:
            if len(search) == 1:
                cdr3_c_end = search[0]

                c_term_nt_trimmed = c_term_nuc[cdr3_c_end * 3:]

                # Add a warning if the detected J match is too far or too short
                if cdr3_c_end > 22:
                    warnings.warn("Warning: germline match \'" + c_term_cdr3_chunk + "\' was found " +
                                  str(search.start()) + " residues past the start of the J, "
                                  "which is an extremely unlikely TCR. ")

                if c <= j_warning_threshold:
                    warnings.warn("Note: while a C-terminal CDR3:J germline match has been found, it was only the "
                                  "string \"" + c_term_cdr3_chunk + "\". ")

            # If no single match found using the non-strict conserved J motif pattern (G..|..G), use the strict GXG
            elif len(search) > 1 and len(c_term_cdr3_chunk) == 1:
                warnings.warn("Note: while a C-terminal CDR3:J germline match has been found, it was only the string "
                              "\"" + c_term_cdr3_chunk + "\", which occurs in two positions. ")

                search2 = [x for x in find_cdr3_c_term(c_term_cdr3_chunk, c_term_amino[:search_len], True)]

                if len(search2) == 1:
                    cdr3_c_end = search2[0]
                    c_term_nt_trimmed = c_term_nuc[cdr3_c_end * 3:]

                else:
                    raise ValueError("CDR3 seemingly deleted up to/past conserved CDR3 junction terminating residue, "
                                     "but multiple motif hits found - unable to locate C terminus of CDR3. ")

            return c_term_nt_trimmed, len(cdr3_cterm_aa) - c

    # Shouldn't be able to get here to throw an error, but just in case
    raise ValueError("Unable to locate C terminus of CDR3 in J gene correctly. Please ensure sequence plausibility. ")


def get_codon_frequencies(path_to_freq_file):
    """
    :param path_to_freq_file: Path to file containing Kazusa-formatted codon usage
    :return: nested counter containing the raw codon frequencies
    """

    codon_usage = coll.defaultdict(nest_counter)
    with open(path_to_freq_file) as in_file:
        for line in in_file:
            cleaned = [x for x in re.sub(r'\(.+?\)', '', line.rstrip()).upper().replace('U', 'T').split(' ') if x]
            if len(cleaned) % 2 != 0:
                raise ValueError("Error in codon usage file - unexpected format.")
            if len(cleaned) == 0:
                continue
            for pair in [x for x in range(len(cleaned)) if x % 2 == 0]:
                codon = cleaned[pair]
                val = cleaned[pair + 1]
                codon_usage[translate_nt(codon)][codon] = float(val)

    return codon_usage


def get_optimal_codons(specified_cu_file, species):
    """
    :param specified_cu_file: Path to file containing Kazusa-formatted codon usage (if specified)
    :param species: human or mouse, for use if a specific absolute path not specified
    :return: dict containing 'best' (most frequent) codon to use per residue
    """

    if specified_cu_file:
        path_to_cu_file = specified_cu_file
    else:
        path_to_cu_file = os.path.join(data_dir, 'kazusa', species + '.txt')

    if not os.path.isfile(path_to_cu_file):
        warnings.warn("Could not find a codon frequency file at this path: " + path_to_cu_file
                      + ". Defaulting to the human table file.")
        path_to_cu_file = os.path.join(data_dir, 'kazusa', 'HUMAN.txt')

    codon_usage = get_codon_frequencies(path_to_cu_file)
    if len(codon_usage) < 20:
        warnings.warn("Warning: incomplete codon usage file input - back translation may fail! ")

    out_dict = coll.defaultdict()
    for residue in codon_usage:
        out_dict[residue] = codon_usage[residue].most_common()[0][0]

    return out_dict


def get_j_motifs(species):
    """
    :param species: upper case string of common species name, referring to a directory in DATA/
    :return: dict of J genes which have a non-canonical (non-phenylalanine) CDR3 terminal residue, and a list of
             J genes whose terminal residue is low confidence (i.e. unable to find a clear FGXG-like motif at all)

    NB: Expects a file in the Data/[species]/ directory called 'J-region-motif.tsv'
    That file should contain a header, with the first 3 columns consisting of:
    J gene/Residue/Confident? (with two later fields of Motif/Position(
    """

    j_file = os.path.join(data_dir, species, 'J-region-motifs.tsv')

    residues = coll.defaultdict()
    low_confidence = []

    with open(j_file, 'r') as in_file:

        line_count = 0
        for line in in_file:
            bits = line.rstrip().split('\t')
            if line_count != 0:
                residues[bits[0]] = bits[1]
                if bits[2] != 'Y':
                    low_confidence.append(bits[0])
            line_count += 1

    return residues, low_confidence


def get_c_motifs(species):
    """
    :param species: upper case string of common species name, referring to a directory in DATA/
    :return: dict of known C genes with  details which allow correct frame inference

    NB: Expects a file in the Data/[species]/ directory called 'C-region-motif.tsv'
    That file should contain a header, with columns consisting of: C gene/Exons/Start motif/Stop codon motif
    """

    c_file = os.path.join(data_dir, species, 'C-region-motifs.tsv')

    constant_motifs = {'start': coll.defaultdict(), 'stop': coll.defaultdict(), 'exons': coll.defaultdict()}

    with open(c_file, 'r') as in_file:

        line_count = 0
        for line in in_file:
            bits = line.rstrip().split('\t')
            if line_count != 0:
                # Record exon configuration
                constant_motifs['exons'][bits[0]] = bits[1]
                # Define a 'start' motif (which gives the correct translation of the start of the C, for frame-finding)
                constant_motifs['start'][bits[0]] = bits[2]
                # If provided, also define a 'stop' motif, which gives a chunk of seq to find in-frame stop codons
                if len(bits) > 3:
                    constant_motifs['stop'][bits[0]] = bits[3]
            line_count += 1

    return constant_motifs


def rev_translate(amino_acid_seq, codon_usage_dict):
    """
    :param amino_acid_seq: An amino acid to convert into nucleotide sequence, using the most common codon
    :param codon_usage_dict: Dict of which codons to use for which amino acids (see get_optimal_codons)
    :return: Corresponding nucleotide sequence
    """

    return ''.join([codon_usage_dict[x] for x in amino_acid_seq])


def get_linker_dict():
    """
    :return: Dictionary of linkers contained in the Data/linkers.tsv file
    """

    if not os.path.isfile(linkers_file):
        raise IOError(linkers_file + " not detected - please check linker file is present and run again. ")

    else:
        linkers = coll.defaultdict()
        with open(linkers_file, 'r') as in_file:
            for line in in_file:
                bits = line.rstrip().split('\t')
                linkers[bits[0]] = bits[1]

        return linkers


def get_linker_seq(linker_text, linker_dict):
    """
    :param linker_text: The text from the Linker field of the input tsv file for bulk stitching (name or custom)
    :param linker_dict: Dict of the known provided linker sequences (from Data/linkers.tsv)
    :return: the corresponding linker sequence, or throw a warning if appropriate
    """

    if linker_text in linker_dict:
        return linker_dict[linker_text]

    # Allows users to input their own custom DNA sequences
    elif dna_check(linker_text):
        if len(linker_text) % 3 != 0:
            warnings.warn("Warning: length of linker sequence \'"
                          + linker_text + "\' is not divisible by 3; "
                          "if this was supposed to be a skip sequence the downstream gene will not be in frame. ")

        return linker_text

    else:
        raise ValueError("Error: " + linker_text +
                         " is not a recognised pre-coded linker sequence and does not seem to be DNA. ")


def dna_check(possible_dna):
    """
    :param possible_dna: A sequence that may or may not be a plausible DNA (translatable!) sequence
    :return: True/False
    """

    return set(possible_dna.upper()).issubset({'A', 'C', 'G', 'T', 'N'})


def tweak_thimble_input(stitch_dict):
    """
    :param stitch_dict: Dictionary produced by stitchr
    :return: Fixed stitchr dict (species capitalised, TCR names blanked)
    """

    stitch_dict['name'] = ''
    for region in ['v', 'j', 'cdr3', 'c']:
        stitch_dict[region] = stitch_dict[region].upper()

    return stitch_dict


def opener(in_file):
    """
    :param in_file: path to file to be opened
    :return: the appropriate file opening command (open or gzip.open)
    """
    if in_file.endswith('.gz'):
        return gzip.open(in_file, 'rt')
    else:
        return open(in_file, 'r')


def translate_nt(nt_seq):
    """
    :param nt_seq: Nucleotide sequence to be translated
    :return: corresponding amino acid sequence
    """
    nt_seq = nt_seq.upper()
    aa_seq = ''
    len_seq = len(nt_seq)
    for i in range(0, len_seq - (len_seq % 3), 3):
        codon = nt_seq[i:i+3]
        if 'N' in codon:
            aa_seq += 'x'
        else:
            try:
                aa_seq += codons[codon]
            except Exception:
                raise IOError("Cannot translate codon: " + codon + ". ")

    return aa_seq


def pad_trans(nt_seq, pad_len):
    """
    :param nt_seq: str of nucleotide sequence to be translated
    :param pad_len: int of translation offset (i.e. amount to pad)
    :return: str of nt_seq translated into AA, after padding with pad_len * N
    """
    return translate_nt('N' * pad_len + nt_seq)


def check_suffix_prefix(five_prime_seq, three_prime_seq):
    """
    :param five_prime_seq: the 5' (or upstream) sequence, in which to look for a suffix
    :param three_prime_seq: the 3' (or downstream) sequence, in which to look for a prefix
    :return: sequence corresponding to the overlap between the five_prime and three_prime seqs
    Based off code from Stack Overflow user 'Mad Physicist' - see https://stackoverflow.com/questions/58598805/
    """

    steps = range(min(len(five_prime_seq), len(three_prime_seq)) - 1, -1, -1)
    return next((three_prime_seq[:n] for n in steps if five_prime_seq[-n:] == three_prime_seq[:n]), '')


def find_v_overlap(v_germline, nt_cdr3):
    """
    :param v_germline: nucleotide sequence of the 5' part of the germline TCR (i.e. L+V)
    :param nt_cdr3: user provided nucleotide sequences covering (and potentially exceeding) the CDR3 junction
    :return: v_germline sequence running up to (but not including) where it overlaps with the nt_cdr3 sequence,
             and the overlapped sequence (to be subtracted from the CDR3 junction nt before checking for J overlap)
    """

    longest_overlap = ''
    index_longest = 0
    # Have to count backwards from the end of the V; only need to go as far as the provided nt_cdr3 length
    # (although some of that will contain J gene residues)
    # Note that the positive check is against the outside chance someone provides an inconsistently short V gene
    for i in [x for x in range(len(v_germline), len(v_germline) - len(nt_cdr3), -1) if x > 0]:

        tmp_fp = v_germline[:i]
        overlap = check_suffix_prefix(tmp_fp, nt_cdr3)
        if len(overlap) > len(longest_overlap):
            longest_overlap = overlap
            index_longest = i

    return v_germline[:index_longest - len(longest_overlap) ], longest_overlap


def find_j_overlap(nt_cdr3, j_germline):
    """
    :param nt_cdr3: user provided nucleotide sequences covering (and potentially exceeding) the CDR3 junction
    :param j_germline: nucleotide sequence of the 3' part of the germline TCR (i.e. J+C)
    :return: j_germline sequence running from (but not including) where it overlaps with the nt_cdr3 sequence
    """

    longest_overlap = ''
    index_longest = 0
    for i in range(len(j_germline)):
        tmp_j = j_germline[i:]
        overlap = check_suffix_prefix(nt_cdr3, tmp_j)
        if len(overlap) > len(longest_overlap):
            longest_overlap = overlap
            index_longest = i

    return j_germline[index_longest + len(longest_overlap):]


def pad_spaces(str_to_pad, len_to_pad):
    """
    :param str_to_pad: str to pad with spaces
    :param len_to_pad: int of length of padding required
    :return: str_to_pad prefixed with len_to_pad number of spaces (for format matching)
    """
    return str_to_pad + ' ' * (len_to_pad - len(str_to_pad))


def output_genbank(sequence_name, full_sequence, description, topology, features, save_dir_path,
                   species, numbering, plot_multi=True, division='SYN',
                   accession='.', version='.', keywords='.', source='.',
                   authors='.', title='Direct Submission', journal='.'):
    """
    Save an annotated GenBank (.gb) file for easy visualisation of DNA features (e.g. stitched TCRs)
    :param sequence_name: str of sequence name, for saving
    :param full_sequence: str of full DNA sequence
    :param description: str of description to include in output file
    :param topology: str of topology ('linear' or 'circular')
    :param features: lists of tuples [(str, str, [list of str], [list of int]], detailing target output features in GB:
        1) name, 2) type (if known), 3) DNA sequence(s) (allowing multi-part features), 4) index location(s) (optional)
    :param save_dir_path: str of path to folder in which genbank files should be saved
    :param species: str of species to go in 'organism' field
    :param numbering: boolean, whether to incrementally number regions of the same type
    :param plot_multi: boolean, whether to plot all instances of multi-mapping sequences given without indexes (def=0)
    :param division: GenBank 'division', as per https://www.ncbi.nlm.nih.gov/genbank/samplerecord/#GenBankDivisionB
    :param accession: optional str field (can be left as default '.')
    :param version: optional str field (can be left as default '.')
    :param keywords: optional str field (can be left as default '.')
    :param source: optional str field (can be left as default '.')
    :param authors: optional str field (can be left as default '.')
    :param title: optional str field (can be left as default 'Direct submission')
    :param journal: optional str field (can be left as default '.')
    :return: [nothing] just saves the appropriate file
    """
    pad_len1 = 12
    pad_len2 = 21

    seq_len = str(len(full_sequence))
    out_str = [pad_spaces('LOCUS', pad_len1) + sequence_name + '         ' + seq_len +
               ' bp    DNA    ' + topology + '    ' + division + '    ' + today('ncbi'),
               pad_spaces('DEFINITION', pad_len1) +
               description,

               pad_spaces('ACCESSION', pad_len1) + accession,
               pad_spaces('VERSION', pad_len1) + version,
               pad_spaces('KEYWORDS', pad_len1) + keywords,
               pad_spaces('SOURCE', pad_len1) + source,
               pad_spaces('  ORGANISM', pad_len1) + species,
               pad_spaces('REFERENCE', pad_len1) + '1 (bases 1 to ' + seq_len + ')',
               pad_spaces('  AUTHORS', pad_len1) + authors,
               pad_spaces('  TITLE', pad_len1) + title,
               pad_spaces('  JOURNAL', pad_len1) + journal,

               pad_spaces('FEATURES', pad_len2) + 'Location/Qualifiers',
               pad_spaces('     source', pad_len2) + '1..' + str(len(full_sequence)),
               pad_spaces('', pad_len2) + '/organism="' + species + '"']

    # Plot the specified input features
    if features:
        upper_seq = full_sequence.upper()
        for feature in features:
            index_count = 1  # Used to number features if numbering

            # Determine feature type
            if feature[1]:
                f_type = feature[1]
                if not f_type.endswith('_'):
                    f_type += '_'
            else:
                f_type = 'misc_feature'

            # Determine feature location
            num_seqs = len(feature[2])
            num_is = len(feature[3])
            if (num_seqs > 0 and num_is > 0) and num_seqs != num_is:
                raise IOError("Provided feature entry has an inappropriate number of index locations detected: "
                              + str(feature))

            for subfeatx in range(num_seqs):
                found = False
                f_name = feature[0].rstrip()
                if numbering:
                    f_name += ' #' + "{:02d}".format(index_count)

                # If index information is provided, use it
                if num_is > 0:
                    index_seq = upper_seq[feature[3][subfeatx]:feature[3][subfeatx] + len(feature[2][subfeatx])]
                    if index_seq == feature[2][subfeatx].upper():
                        out_str.append(genbank_write_feature(f_name, feature[3][subfeatx],
                                                                 len(feature[2][subfeatx]), f_type))
                        found = True

                # Otherwise look for sequence matches
                if not found:
                    hits = [x for x in find_all_substr(upper_seq, feature[2][subfeatx].upper())]
                    if len(hits) == 1:
                        out_str.append(genbank_write_feature(f_name, hits[0],
                                                             len(feature[2][subfeatx]), f_type))
                    elif plot_multi:
                        for hit in hits:
                            out_str.append(genbank_write_feature(f_name + ' (multiple matches)', hit,
                                                                 len(feature[2][subfeatx]), f_type))

                index_count += 1

    out_str.append('ORIGIN')
    out_str.append(ncbi_format_dna(full_sequence, pad_len1))
    out_str.append('//')

    with open(os.path.join(save_dir_path, sequence_name) + '.gb', 'w') as out_file:
        out_file.write('\n'.join(out_str))


def find_all_substr(substr, full_seq):
    """
    :param substr: str, substring to look for in ...
    :param full_seq: str, longer string in which to look for instances of the provided substring
    :yield: indexes of all matches
    NB: case sensitive
    """
    match = full_seq.find(substr)
    while match != -1:
        yield match
        match = full_seq.find(substr, match + 1)


def genbank_write_feature(feat_name, feat_loc, feat_len, feat_type='misc_feature', space_len=21):
    """
    :param feat_name: str, complete name of DNA feature (including any numbering / qualifiers etc)
    :param feat_loc: str, describing 1-indexed position range of the feature in format 'X..Y'
    :param feat_len: int, length of the DNA feature
    :param feat_type: str, type of DNA feature if known
    :param space_len: int, length of spacer used to left indent feature text in GenBank file
    :return: str of formatted GenBank entry for a given feature
    """

    feat_str = (pad_spaces('     ' + feat_type, space_len)
                + str(feat_loc + 1) + '..' + str(feat_loc + feat_len) + '\n' +
                pad_spaces('', space_len) + '/label="' + feat_name + '"'
                )
    return feat_str


def ncbi_format_dna(in_str, pad_len):
    """
    :param in_str: str of DNA sequence
    :param pad_len: int of padding required
    :return: str of DNA sequence in NCBI format (indented/space-separated 10-mers)
    """
    dna_text = ''
    dna_chunks = [in_str[x:x+10] for x in range(0, len(in_str), 10)]
    line_chunks = [dna_chunks[x:x+6] for x in range(0, int(len(in_str)/6), 6)]
    pos = 1
    for line in line_chunks:
        if line:
            dna_text += ' '.join([num_pad(pos, pad_len)] + line) + '\n'
            pos += 60
        else:
            break

    return dna_text


def num_pad(number, len_to_pad):
    """
    :param number: int describing a number to be converted to str and prefixed with spaces
    :param len_to_pad: int of number of spaces to pad
    :return: str of 'number' prefixed with len_to_pad number of spaces (for format matching)
    """
    str_num = str(number)
    return ' ' * (len_to_pad - len(str_num)) + str_num


codons = {'AAA': 'K', 'AAC': 'N', 'AAG': 'K', 'AAT': 'N',
          'ACA': 'T', 'ACC': 'T', 'ACG': 'T', 'ACT': 'T',
          'AGA': 'R', 'AGC': 'S', 'AGG': 'R', 'AGT': 'S',
          'ATA': 'I', 'ATC': 'I', 'ATG': 'M', 'ATT': 'I',
          'CAA': 'Q', 'CAC': 'H', 'CAG': 'Q', 'CAT': 'H',
          'CCA': 'P', 'CCC': 'P', 'CCG': 'P', 'CCT': 'P',
          'CGA': 'R', 'CGC': 'R', 'CGG': 'R', 'CGT': 'R',
          'CTA': 'L', 'CTC': 'L', 'CTG': 'L', 'CTT': 'L',
          'GAA': 'E', 'GAC': 'D', 'GAG': 'E', 'GAT': 'D',
          'GCA': 'A', 'GCC': 'A', 'GCG': 'A', 'GCT': 'A',
          'GGA': 'G', 'GGC': 'G', 'GGG': 'G', 'GGT': 'G',
          'GTA': 'V', 'GTC': 'V', 'GTG': 'V', 'GTT': 'V',
          'TAA': '*', 'TAC': 'Y', 'TAG': '*', 'TAT': 'Y',
          'TCA': 'S', 'TCC': 'S', 'TCG': 'S', 'TCT': 'S',
          'TGA': '*', 'TGC': 'C', 'TGG': 'W', 'TGT': 'C',
          'TTA': 'L', 'TTC': 'F', 'TTG': 'L', 'TTT': 'F'}

regions = {'l': 'LEADER',
           'v': 'VARIABLE',
           'j': 'JOINING',
           'c': 'CONSTANT'
           }

spacer = ' ' * 12

class GetCitation(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        print(get_citation())
        sys.exit(0)


def get_package_version():
    """
    :return: str of the currently installed Stitchr module
    """
    from importlib.metadata import version
    return version('Stitchr')


def get_metadata_dict(species, script, script_version):
    """
    :param species: str, species name
    :param script: str, name of script calling the function
    :param script_version: str, the __version__ of said script
    :return: nested dict of the reference and run information
    """

    meta_dict = {'reference': coll.defaultdict(str), 'stitchr': coll.defaultdict(str)}
    sp_meta_file = data_dir + '/' + species + '/data-production-date.tsv'

    # Get the germline reference details from the 'data-production-date' file
    with open(sp_meta_file, 'r') as in_file:
        for line in in_file:
            bits = line.rstrip().split('\t')
            if len(bits) != 2:
                raise IOError("Incorrect number of fields in reference file line! Expecting two, got: " + str(bits))
            meta_dict['reference'][bits[0]] = bits[1]

    # Then populate the details relating to the running of the scripts themselves
    meta_dict['stitchr']['date_run'] = today()
    meta_dict['stitchr']['script'] = script
    meta_dict['stitchr']['script_version'] = script_version
    meta_dict['stitchr']['package_version'] = get_package_version()
    meta_dict['stitchr']['command'] = ' '.join(sys.argv[:])

    return meta_dict


def get_metadata_text(species, script, script_version):
    """
    :param species: str, name of species
    :param script: str, name of script calling the function
    :param script_version: str, the __version__ of said script
    :return: str of a text description of the data used to stitch this specific TCR
    """

    meta = get_metadata_dict(species, script, script_version)
    return ("Sequence generated using " + script + " (v" + script_version + ", package version " +
            get_package_version() + "), on " + meta['stitchr']['date_run'] + ", using the " + species +
            " " + meta['reference']['reference'] + " reference (release " + meta['reference']['release'] +
            ", downloaded on " + meta['reference']['last_run'] + "). ")


def get_citation():
    """
    print out paper and data details for proper citation and reproducibility purposes
    """

    citation_str = ('\nPaper citation details:\n\tJames M Heather, Matthew J Spindler, Marta Herrero Alonso, Yifang '
                    'Ivana Shui, David G Millar, David S Johnson, Mark Cobbold, & Aaron N Hata. '
                    'Stitchr: stitching coding TCR nucleotide sequences from V/J/CDR3 information, Nucleic Acids '
                    'Research, Volume 50, Issue 12, 8 July 2022, Page e68, https://doi.org/10.1093/nar/gkac190\n\n')

    from . import stitchr as st
    from . import thimble as th
    from . import gui_stitchr as gui

    citation_str += ('Version numbers:\n\twhole package:\t' + get_package_version() +
                     '\n\tstitchr:\t' + st.__version__ +
                     '\n\tthimble:\t' + th.__version__ +
                     '\n\tgui_stitchr:\t' + gui.__version__ + '\n\n')

    # Loop across all installed species data:
    try:
        species_used = find_species_covered()
        for sp in species_used:
            sp_dat = get_metadata_dict(sp, 'stitchr', __version__)
            citation_str += (sp + ' reference data details:\n\t' + sp_dat['reference']['reference'] + ' v:\t' +
                             sp_dat['reference']['release'] +
                             '\n\tDownloaded on:\t' + sp_dat['reference']['last_run'] +
                             '\n\tUsing script:\t' + sp_dat['reference']['script_used'] +
                             ' (v ' + sp_dat['reference']['version_used'] + ')\n')
    except Exception:
        citation_str += '(No species data detected in the data directory.)\n'

    # Check for novel allele additions
    na_prefix = 'novel-TCR-alleles-'
    novel_check =  [x for x in os.listdir(data_dir) if x.startswith(na_prefix)
                    and x.endswith('.fasta')]
    if novel_check:
        novel_check.sort()
        built_on, novel_v, genedb_v = (novel_check[-1].replace(na_prefix, '').replace('.tsv', '').
                                       replace('GENEDB-', '').split('_'))
        citation_str += ('Novel human TCR alleles (from https://github.com/JamieHeather/novel-tcr-alleles) detected:\n'
                         '\tBuilt on:\t' + built_on +
                         '\n\tNovel code v:\t' + novel_v +
                         '\n\tIMGT/GENE-DB v:\t' + genedb_v + '\n')

    return citation_str


def get_output_name(stitch_dict):
    """
    :param stitch_dict: dict produced from the output of running stitch(), which also contains the input arg dict
    :return: str of a name for use in naming output files
    """
    # Preferentially use provided names
    if stitch_dict['in']['name']:
        name = stitch_dict['in']['name']
    else:
        name = ('stitchr-TCR_' +
                stitch_dict['used']['v'].replace('*', '-') + '_' +
                stitch_dict['used']['j'].replace('*', '-') + '_' +
                stitch_dict['used']['cdr3'])
    return name


def main():
    print("Please use the appropriate 'stitchr', 'thimble', 'gui_stitchr' or 'stitchrdl' command.")

# TODO make text outputs referencing IMGT instead reference whatever germline set source was used



# -*- coding: utf-8 -*-

"""
stitchrdl.py

Download IMGT/GENE-DB data for stitching, for a given species
"""
from . import stitchrfunctions as fxn
import argparse
import os
import subprocess
import sys
import shutil

# Make sure IMGTgeneDL is installed
try:
    import IMGTgeneDL
except (ImportError, ModuleNotFoundError) as err:
    print("IMGTgeneDL is required to run this script (pip install IMGTgeneDL)")
    sys.exit()


__version__ = '0.3.0'
__author__ = 'Jamie Heather'
__email__ = 'jheather@mgh.harvard.edu'

sys.tracebacklimit = 0  # comment when debugging

def args():
    """
    args(): Obtains command line arguments which dictate the script's behaviour
    """

    # Help flag
    parser = argparse.ArgumentParser(
        description="stitchrdl v" + str(__version__) + '\n' + ": Process input data for use with stitchr. \n"
                    "See https://github.com/JamieHeather/stitchr and https://doi.org/10.1093/nar/gkac190.")

    # Input and output options
    parser.add_argument('-s', '--species', required=False, type=str, default='HUMAN',
                        help="Species (common name). Optional. Default = HUMAN")

    parser.add_argument('--version', action='version', version=__version__,
                        help="Print current stitchr version.")

    parser.add_argument('-fa', '--fasta', required=False, type=str, default='',
                        help="Provide a path to a FASTA file to add sequences to the additional-genes.fasta reference.")

    parser.add_argument('-na', '--novel_alleles', required=False, action='store_true', default=False,
                        help="Option to download collated published novel TCR alleles "
                             "to the additional-genes.fasta reference.")

    parser.add_argument('-ns', '--novel_studies_threshold', required=False, type=int, default=2,
                        help="If using the -na/--novel_alleles tag, this optional field allows specifying the number of"
                             "studies an allele must be found in to be included. Default = 2.")

    parser.add_argument('-nd', '--novel_donors_threshold', required=False, type=int, default=3,
                        help="If using the -na/--novel_alleles tag, this optional field allows specifying the number of"
                             "donors (across all studies) an allele must be found in to be included. Default = 3.")

    parser.add_argument('--cite', action=fxn.GetCitation, help="Print citation details.", nargs=0)

    return parser.parse_args()


region_key = {'L-': 'LEADER', 'V-': 'VARIABLE', 'J-': 'JOINING', 'EX': 'CONSTANT', 'CH': 'CONSTANT',
              'V': 'VARIABLE', 'J': 'JOINING', 'C': 'CONSTANT'}


def infer_genetype_genename(gene_name, gene_seq):
    """
    :param gene_name: str of gene name as supplied
    :param gene_seq: str of DNA sequence
    :return: inferred gene region
    NB for 'V' genes sequences <100 nt are presumed to be leaders, >100 nt presumed to be V-REGIONS
    """
    if gene_name[:2] in ['TR', 'IG'] and gene_name[3] in ['V', 'J', 'C']:
        if gene_name[3] == 'V':
            if len(gene_seq) < 100:
                print("Based off FASTA header details and gene length, '" + gene_name +
                      "' was presumed to be a leader sequence.")
                return 'LEADER'
            else:
                print("Based off FASTA header details and gene length, '" + gene_name +
                      "' was presumed to be a variable gene sequence.")
                return 'VARIABLE'
        else:
            return region_key[gene_name[3]]
    else:
        raise IOError("Cannot infer gene type for gene '" + gene_name + "' - please use full IMGT style format. ")


def infer_genetype_whole(fasta_header_bits, fasta_seq):
    """
    :param fasta_header_bits: 16-item list of strs, detailing the contents of an IMGT-style FASTA header
    :param fasta_seq: string of fasta read, to help distinguish V-REGIONs from leader sequences
    :return: str of 'L', 'V', 'J', or 'C' if a gene type can be inferred, or a null value if not
    """
    gene, allele = fasta_header_bits[1].split('*')

    # Preferentially go off the specific region field (as this is the most precise)
    region = fasta_header_bits[4][:2]
    if region:
        if region in region_key:
            return region_key[region]
    # Otherwise try to use the gene name (and sequence
    return infer_genetype_genename(gene, fasta_seq)


def gene_allele_check(given_id):
    if '*' in given_id:
        gene, allele = given_id.split('*')
        if gene and allele:
            return True
        else:
            return False


def add_from_fasta(path_to_fasta):
    """
    :param path_to_fasta: str detailing path to FASTA file containing sequences to add to the additional gene file
    Saves the additional gene FASTA file with new reads from the provided FASTA
    NB: exact ID matches to an existing entry will over-write
    """
    # Read in both existing additional gene file, and the specified sequences to add
    current_fa = fxn.fasta_to_dict(fxn.additional_genes_file)
    to_add_fa = fxn.fasta_to_dict(path_to_fasta)

    # Then go through the new file and add the reads in
    for new_entry in to_add_fa:
        # Check if entry meets the required criteria
        bits = new_entry.split('|')
        imgt_format, gene_allele, inferrable_type, stated_type = [False] * 4
        inferrable_type, stated_type = False, False
        if len(bits) == 16:
            imgt_format = True
            if gene_allele_check(bits[1]):
                gene, allele = bits[1].split('*')
                gene_allele = True
                if len(gene) > 3:
                    if gene[3] in ['V', 'J', 'C'] or bits[4][:2] in ['L-', 'V-', 'J-', 'EX', 'CH']:
                        inferrable_type = True
                if bits[15] in ['~LEADER', '~VARIABLE', '~JOINING', '~CONSTANT']:
                    stated_type = True
        elif len(bits) > 1:
            raise IOError("Cannot automatically read in sequence (need either just sequence ID or full IMGT format) for "
                          + new_entry)

        if not gene_allele and not gene_allele_check(new_entry):
            raise IOError("Failed to read in " + new_entry + ".\n"
                          "Input FASTA entries must have a proper asterisk-delimited 'gene*allele' field (e.g. "
                          "TRAV1-1*01), noting that any arbitrary string can be used in either section. ")

        if imgt_format and gene_allele and stated_type:
            current_fa[new_entry] = to_add_fa[new_entry]
        elif imgt_format and gene_allele and inferrable_type:
            inferred_type = infer_genetype_whole(bits[1], to_add_fa[new_entry])
            current_fa[new_entry[:new_entry.rfind('|')+1] + '~' + inferred_type] = to_add_fa[new_entry]
        elif not imgt_format:
            inferred_type = infer_genetype_genename(new_entry, to_add_fa[new_entry])
            new_id = '|'.join(['stitchrdl-added', new_entry] + [' '] * 14)
            new_id = new_id[:new_id.rfind('|')+1] + '~' + inferred_type
            current_fa[new_id] = to_add_fa[new_entry]

        else:
            raise IOError("Failed to automatically read in entry for '" + new_entry +
                          "' - please try full formatting and try again. ")

    # Then archive the current version of the additional-genes file,
    # to minimise the chance of losing sequences (and thus losing reproducibility)...
    shutil.move(fxn.additional_genes_file,
                os.path.join(fxn.data_dir, 'archived-' + fxn.today() + '-additional-genes.fasta'))

    # ... before writing out the modified additional-genes.fasta file in its place
    with open(fxn.additional_genes_file, 'w') as out_file:
        for entry in current_fa:
            out_file.write(fxn.fastafy(entry, current_fa[entry]))


def download_novel_alleles(study_threshold, donor_threshold):
    """
    :param study_threshold: int of # of studies a given allele must be found in to be retained
    :param donor_threshold: int of # of donors a given allele must be found in to be retained, summed across all studies
    This function grabs novel alleles from the repo where I collate them, and reads them into the additional-genes file
    """

    import pandas as pd
    import requests

    novel_repo_url = 'https://api.github.com/repos/JamieHeather/novel-tcr-alleles/contents/'
    summary_prefix = 'novel-TCR-alleles-'

    # Identify the current novel-tcr-allele file
    response = requests.get(novel_repo_url, headers={})
    novel_file_name = ''
    if response.status_code == 200:
        files = response.json()
        matching_files = [x for x in files if x['name'].startswith(summary_prefix)]
        if len(matching_files) == 1:
            novel_file_name = matching_files[0]['name']
            novel_file_url = matching_files[0]['download_url']

    if not novel_file_name:
        raise IOError("Unable to locate a suitable summary novel TCR allele file name in the GitHub repository. ")

    # If found, download it
    tsv_path = os.path.join(fxn.data_dir, novel_file_name)
    response = requests.get(novel_file_url, headers={})

    if response.status_code == 200:
        with open(tsv_path, 'wb') as out_file:
            out_file.write(response.content)
    else:
        raise IOError("Failed to download the novel TCR allele data. ")

    # Then read in and whittle down to those entries that meet select criteria
    novel = pd.read_csv(tsv_path, sep='\t')
    novel_out_fasta = os.path.join(fxn.data_dir, novel_file_name.replace('.tsv', '.fasta'))
    with open(novel_out_fasta, 'w') as out_file:
        for row in novel.index:
            row_bits = novel.loc[row]
            if pd.isna(row_bits['Notes']):
                notes = ''
            else:
                notes = str(row_bits['Notes'])
            func = '?'

            if 'Stop codon' in notes:
                func = 'P'

            # Disregard alleles found in fewer than the threshold # of studies
            if row_bits['Number-Datasets-In'] < study_threshold:
                continue
            # Disregard alleles found in fewer than the threshold # of donors
            elif row_bits['Number-Donors-In'] < donor_threshold:
                continue
            # If it's already added to IMGT, skip
            elif isinstance(row_bits['IMGT-ID'], str):
                continue
            # Same for if it's a shorter version
            elif 'Shorter version of IMGT' in notes:
                continue

            # Determine whether the allele has a valid name
            allele_id = ''
            if row_bits['Standard-ID'].startswith('TR'):
                allele_id = row_bits['Standard-ID']
            else:
                # If not, borrow one of the names from the paper(s) where it was discovered
                studies = [x for x in novel if '-Name' in x]
                for study in studies:
                    if not pd.isna(row_bits[study]):
                        allele_id = row_bits[study] + '-' + study.replace('-Name', '')
                        if allele_id[:2] != 'TR' or '*' not in allele_id:
                            allele_id = row_bits['Gene'] + '*' + allele_id

            if not allele_id:
                raise IOError("Unable to determine a possible allele ID for sequence with data:\n" + str(row_bits))

            else:
                header = '|'.join([novel_file_name.replace('.tsv', ''), allele_id, 'Homo sapiens', func,
                                   '', '', str(len(row_bits['Ungapped-Sequence'])) + ' nt',
                                   '', '', '', '', '', '', '', notes, '~' + region_key[allele_id[3]]])
                out_file.write(fxn.fastafy(header, row_bits['Ungapped-Sequence']))

    # Having written out the selected novel alleles FASTA, add that to the additional-genes file
    add_from_fasta(novel_out_fasta)
    print("Successfully added novel alleles from " + novel_file_name.replace('.tsv', ''))


def main():

    in_args = vars(args())
    alt_text_suffix = " to the additional-genes.fasta file, in the Stitchr data directory. "

    if in_args['fasta']:
        print("Reading in sequences from " + in_args['fasta'] + alt_text_suffix)
        add_from_fasta(in_args['fasta'])

    elif in_args['novel_alleles']:
        print("Downloading published novel human TCR alleles from GitHub" + alt_text_suffix)
        download_novel_alleles(in_args['novel_studies_threshold'], in_args['novel_donors_threshold'])

    else:
        # Default behaviour, download from IMGT/GENE-DB via IMGTgeneDL
        species = in_args['species'].upper()

        try:
            cmd = 'IMGTgeneDL -m stitchr -n -s ' + species
            subprocess.call(cmd, shell=True)
        except Exception:
            print("Failed to download sequences from IMGTgeneDL.")

        # Assuming that made the folder appropriately
        if species in os.listdir(os.getcwd()):
            # Delete older entries
            if species in os.listdir(fxn.data_dir):
                print("Note: a directory for this species exists: overwriting.")
                shutil.rmtree(os.path.join(fxn.data_dir, species))
            shutil.move(os.path.join(os.getcwd(), species), os.path.join(fxn.data_dir, species))

        else:
            raise IOError("Failed to produce the expected species\' output directory. ")



# -*- coding: utf-8 -*-

"""
thimble.py

Runs stiTChR in a high-throughput manner, on a tab separated input file.

No fancy backronym or contrived silent capitals - it just helps you stitch faster.

"""


from . import stitchrfunctions as fxn
from . import stitchr as st
import warnings
import argparse
import os
import sys
import collections as coll
from time import time

__version__ = '1.3.0'
__author__ = 'Jamie Heather'
__email__ = 'jheather@mgh.harvard.edu'

sys.tracebacklimit = 0  # comment when debugging


def args():
    """
    args(): Obtains command line arguments which dictate the script's behaviour
    """

    # Help flag
    parser = argparse.ArgumentParser(description="thimble v" + str(__version__) + '\n' +
                                                 ": Run stiTChR on multiple and paired TCRs")

    # Input and output options
    parser.add_argument('-in', '--in_file', required=True, type=str,
                        help="Tab-separated input file. Must fit the format of the example file. Required.")

    parser.add_argument('-o', '--out_file', required=True, type=str,
                        help="Path/file name for output tsv file. Required.")

    parser.add_argument('-s', '--species', required=False, type=str, default='HUMAN',
                        help="Species. Optional. Default = 'HUMAN', options depend on directories in Data directory.")

    parser.add_argument('-m', '--mode', required=False, type=str, default='TSV',
                        help="Standard out output mode. Options are 'TSV' (default), 'GB', or 'JSON'.")

    parser.add_argument('-ol', '--only_linked', required=False, action='store_true', default=False,
                        help="Optional flag to speed up thimble by only translating linked/paired chain sequences.")

    parser.add_argument('-xg', '--extra_genes', action='store_true', required=False, default=False,
                        help="Optional flag to use additional (non-database/-natural) sequences in "
                             "the \'additional-genes.fasta\' file.")

    parser.add_argument('-sl', '--seamless', action='store_true', required=False, default=False,
                        help="Optional flag to integrate known nucleotide sequences seamlessly. \nNB: "
                             "nucleotide sequences covering the CDR3 junction with additional V gene context required.")

    parser.add_argument('-r', '--receptor', required=False, type=str,
                        help="Receptor to stitch, alpha/beta or gamma/delta TCR. Optional. Give double or single digits"
                             " (e.g. 'ab', 'gd', 'd', 'b').")

    parser.add_argument('-p', '--preferred_alleles_path', required=False, type=str, default='',
                        help="Path to a file of preferred alleles to use when no allele specified (instead of *01). "
                             "Optional.")

    parser.add_argument('-cu', '--codon_usage', required=False, type=str,
                        help="Path to a file of Kazusa-formatted codon usage frequencies. Optional.")

    parser.add_argument('-jt', '--j_warning_threshold', required=False, type=int, default=3,
                        help="J gene substring length warning threshold. Default = 3. "
                             "Decrease to get fewer notes on short J matches. Optional.")

    parser.add_argument('--version', action='version', version=__version__,
                        help="Print current thimble version.")

    parser.add_argument('--cite', action=fxn.GetCitation, help="Print citation details.", nargs=0)

    return parser.parse_args()


def locus_to_trx(trx_str):
    """
    :param trx_str: str containing (a) reference to a specific locus/loci (e.g. TRA/TRB or TRG/TRD)
    :return: same string, but with those loci specific IDs replaced with generic TR1/TR2 strings
    """
    return trx_str.replace('TRA', 'TR1').replace('TRB', 'TR2').replace('TRG', 'TR1').replace('TRD', 'TR2')


def populate_blanks(chain_dict, necessary_fields):
    """
    :param chain_dict: dictionary containing the stitchr fields for a given chain's recombination
    :param necessary_fields: list of fields that the tcr dictionary requires to pass the downstream steps
    :return: chain_dict with blank entries for each of the necessary_fields that it was missing
    """
    for nf in necessary_fields:
        if nf not in chain_dict:
            chain_dict[nf] = ''

    return chain_dict


in_headers = {'TRA/TRB': ['TCR_name', 'TRAV', 'TRAJ', 'TRA_CDR3', 'TRBV', 'TRBJ', 'TRB_CDR3',
                          'TRAC', 'TRBC', 'TRA_leader', 'TRB_leader', 'Linker', 'Link_order',
                          'TRA_5_prime_seq', 'TRA_3_prime_seq', 'TRB_5_prime_seq', 'TRB_3_prime_seq'],
              'TRG/TRD': ['TCR_name', 'TRGV', 'TRGJ', 'TRG_CDR3', 'TRDV', 'TRDJ', 'TRD_CDR3',
                          'TRGC', 'TRDC', 'TRG_leader', 'TRD_leader', 'Linker', 'Link_order',
                          'TRG_5_prime_seq', 'TRG_3_prime_seq', 'TRD_5_prime_seq', 'TRD_3_prime_seq']}

convert_fields = {'TRAV': 'v', 'TRAJ': 'j', 'TRA_CDR3': 'cdr3', 'TRBV': 'v', 'TRBJ': 'j', 'TRB_CDR3': 'cdr3',
                  'TRAC': 'c', 'TRBC': 'c', 'TRA_leader': 'l', 'TRB_leader': 'l', 'TCR_name': 'name',
                  'TRA_5_prime_seq': '5_prime_seq', 'TRA_3_prime_seq': '3_prime_seq',
                  'TRB_5_prime_seq': '5_prime_seq', 'TRB_3_prime_seq': '3_prime_seq',
                  'TRGV': 'v', 'TRGJ': 'j', 'TRG_CDR3': 'cdr3', 'TRDV': 'v', 'TRDJ': 'j', 'TRD_CDR3': 'cdr3',
                  'TRGC': 'c', 'TRDC': 'c', 'TRG_leader': 'l', 'TRD_leader': 'l',
                  'TRG_5_prime_seq': '5_prime_seq', 'TRG_3_prime_seq': '3_prime_seq',
                  'TRD_5_prime_seq': '5_prime_seq', 'TRD_3_prime_seq': '3_prime_seq'}

pre_stitch_list_fields = ['name', 'v', 'j', 'c', 'cdr3', 'l', '5_prime_seq', '3_prime_seq']
post_stitch_list_fields = ['_name', 'V', 'J', 'C', '_CDR3', '_leader', '_5_prime_seq', '_3_prime_seq']

out_headers = {'TRA/TRB': ['TCR_name', 'TRA_nt', 'TRB_nt', 'TRA_aa', 'TRB_aa',
                           'TRAV', 'TRAJ', 'TRA_CDR3', 'TRBV', 'TRBJ', 'TRB_CDR3',
                           'TRAC', 'TRBC', 'TRA_leader', 'TRB_leader', 'Linker', 'Link_order',
                           'TRA_5_prime_seq', 'TRA_3_prime_seq', 'TRB_5_prime_seq', 'TRB_3_prime_seq',
                           'Linked_nt', 'Linked_aa', 'Warnings/Errors'],
               'TRG/TRD': ['TCR_name', 'TRG_nt', 'TRD_nt', 'TRG_aa', 'TRD_aa',
                           'TRGV', 'TRGJ', 'TRG_CDR3', 'TRDV', 'TRDJ', 'TRD_CDR3',
                           'TRGC', 'TRDC', 'TRG_leader', 'TRD_leader', 'Linker', 'Link_order',
                           'TRG_5_prime_seq', 'TRG_3_prime_seq', 'TRD_5_prime_seq', 'TRD_3_prime_seq',
                           'Linked_nt', 'Linked_aa', 'Warnings/Errors']}


def main():

    # Get input arguments, get the required data read in
    input_args = vars(args())
    start = time()

    # Get species, in order of command line arg > inferred from input file name > default human
    if input_args['species']:
        if input_args['species'].upper() in fxn.find_species_covered():
            species = input_args['species'].upper()
        else:
            raise IOError("No data available for requested species: " + input_args['species'])
    else:
        species_inference = fxn.infer_species(input_args['in_file'])
        if species_inference:
            species = species_inference
        else:
            species = 'HUMAN'

    codons = fxn.get_optimal_codons(input_args['codon_usage'], species)
    linker_dict = fxn.get_linker_dict()

    # Fetch required motifs from the species folder in the data directory
    j_res, low_conf_js = fxn.get_j_motifs(species)
    c_res = fxn.get_c_motifs(species)

    tcr_dat = {}
    tcr_functionality = {}
    preferences = {}

    # Figure out whether a/b or g/d
    if input_args['receptor']:
        input_receptor = input_args['receptor'].upper()
        if ('A' in input_receptor or 'B' in input_receptor) and not ('G' in input_receptor or 'D' in input_receptor):
            receptor = 'TRA/TRB'
        elif ('G' in input_receptor or 'D' in input_receptor) and not ('A' in input_receptor or 'B' in input_receptor):
            receptor = 'TRG/TRD'
        else:
            raise IOError("Unable to determine receptor from '-r' command string: " + input_receptor + ". ")

    else:  # If not explicitly provided, infer from input TSV headers
        with fxn.opener(input_args['in_file']) as in_file:
            for line in in_file:
                if 'TRAV' in line and 'TRGV' not in line:
                    receptor = 'TRA/TRB'
                elif 'TRGV' in line and 'TRAV' not in line:
                    receptor = 'TRG/TRD'
                else:
                    raise IOError("Unable to determine receptor from input file header, please check template. ")
                break

    # Define the individual receptors (chains or loci, i.e. TRA and TRB or TRG and TRD) in play
    r1, r2 = receptor.split('/')
    for c in [r1, r2]:

        tmp_tcr_dat, tmp_functionality, partial = fxn.get_ref_data(c, st.gene_types, species)
        tcr_dat[c] = tmp_tcr_dat
        tcr_functionality[c] = tmp_functionality

        if 'extra_genes' in input_args:
            if input_args['extra_genes']:
                tcr_dat[c], tcr_functionality[c] = fxn.get_additional_genes(tcr_dat[c], tcr_functionality[c])
                input_args['skip_c_checks'] = True
            else:
                input_args['skip_c_checks'] = False

        # Allow for provision of preferred alleles
        if input_args['preferred_alleles_path']:
            preferences[c] = fxn.get_preferred_alleles(input_args['preferred_alleles_path'], list(fxn.regions.values()),
                                                       tcr_dat[c], partial, c)
        else:
            preferences[c] = ''

    if not input_args['out_file'].lower().endswith('.tsv'):
        input_args['out_file'] += '.tsv'

    # If using certain file output options, sort out the environment to cope with their specific outputs
    input_args['mode'] = input_args['mode'].upper()
    if input_args['mode'] in ['GB', 'JSON']:

        base_description = fxn.get_metadata_text(input_args['species'], 'thimble', __version__)

        base_description += ("Produced using the thimble command: '" + ' '.join(sys.argv[:]) + "'. ")

        if input_args['extra_genes']:
            base_description += ("Note that this was run using the 'extra genes' flag, and thus additional "
                                 "non-reference sequences may have been incorporated. ")

        input_args['out_folder'] = input_args['out_file'][:-4]
        if not os.path.exists(input_args['out_folder']):
            os.makedirs(input_args['out_folder'])

        if input_args['mode'] == 'GB':
            # Also create the individual subdirectories (unless 'only_linked' option is selected)
            subdirs = ['linked']
            if not input_args['only_linked']:
                subdirs += [r1, r2]
            for subdir in subdirs:
                subdir_to_make = os.path.join(input_args['out_folder'], subdir)
                if not os.path.exists(subdir_to_make):
                    os.makedirs(subdir_to_make)

        elif input_args['mode'] == 'JSON':
            import json
    else:
        input_args['out_folder'] = os.getcwd()

    # Then go through in file and stitch each TCR on each line
    if not os.path.isfile(input_args['in_file']):
        raise IOError(input_args['in_file'] + " not detected - please check and specify in file again.")

    with fxn.opener(input_args['in_file']) as in_file:

        line_count = 1
        tcr_count = 1
        out_data = ['\t'.join(out_headers[receptor])]

        for line in in_file:

            bits = line.replace('\n', '').replace('\r', '').replace('\"', '').split('\t')

            # Check there's the right number of columns (with the right headers)
            if line_count == 1:
                if bits != in_headers[receptor]:
                    raise ValueError("Headers in input file don't match the expectations - please check template.")

            else:
                # Generate a dict per line ...
                line_sorted_row_bits = {x: '' for x in out_headers[receptor]}
                entry_line_warnings = []

                if len(bits) != len(in_headers[receptor]):  # Pad entries from TSVs where whitespace on right is missing
                    bits += [''] * (len(in_headers[receptor]) - len(bits))

                # Allow for users to specify multiple options per line, to allow the quick stitching of related TCRs
                multiple_options = []
                for i in range(len(in_headers[receptor])):
                    line_sorted_row_bits[in_headers[receptor][i]] = bits[i]
                    if ',' in bits[i] or '%' in bits[i]:
                        multiple_options.append(i)

                # Make sure there's a (usable) name
                if line_sorted_row_bits['TCR_name']:
                    line_sorted_row_bits['TCR_name'] = line_sorted_row_bits['TCR_name'].strip()
                else:
                    line_sorted_row_bits['TCR_name'] = 'TCR-' + "{:05d}".format(tcr_count)

                tcr_lines = [line_sorted_row_bits]
                if multiple_options:
                    # Iterate over all possible combinations for multiply defined options
                    for multi_var in multiple_options:
                        updated_lines = []
                        var_name = in_headers[receptor][multi_var]
                        region = fxn.regions[var_name[3:].replace('_', '')[0].lower()]

                        # A manually entered list
                        if ',' in bits[multi_var]:
                            var_options = bits[multi_var].split(',')

                        # If not a list, it should be a wild card, but only for specified fields
                        elif var_name[3:] not in ['V', 'J', 'C', '_leader']:
                            entry_line_warnings.append("% wild cards only allowed in V/J/C/leader fields. ")
                            break

                        # Wild card just all alleles for a certain gene
                        elif '*%' in bits[multi_var]:
                            gene = bits[multi_var].split('*')[0]
                            if gene in tcr_dat[var_name[:3]][region]:
                                var_options = [gene + '*' + x for x in tcr_dat[var_name[:3]][region][gene].keys()]
                            else:
                                entry_line_warnings.append("Specified multiple entry option gene '" + gene + "' is not "
                                                           "found in the " + region + " region data. ")
                                break

                        # Or wild card all genes/allele combinations for a given region (AKA throw it all at the wall)
                        elif bits[multi_var] == '%':
                            var_options = []
                            for g in tcr_dat[var_name[:3]][region]:
                                for a in tcr_dat[var_name[:3]][region][g]:
                                    var_options.append(g + '*' + a)

                        else:
                            entry_line_warnings.append("Unexpected multi-option field entry detected: " +
                                                       bits[multi_var] + '. ')
                            break

                        # Then loop through those options filling out all specified combinations to stitch
                        for vo in var_options:
                            for out_line in tcr_lines:
                                updated_lines.append(dict(out_line))
                                updated_lines[-1][var_name] = vo
                                updated_lines[-1]['TCR_name'] = updated_lines[-1]['TCR_name'] + '-' + \
                                                                str(var_options.index(vo))
                        tcr_lines = updated_lines

                if entry_line_warnings:
                    tcr_lines[0]['Warnings/Errors'] = ''.join(entry_line_warnings)
                    # out_data.append('\t'.join([tcr_lines[0][x] for x in out_headers[receptor]]))
                    print(2, out_data)
                    continue

                for sorted_row_bits in tcr_lines:

                    thimble_dict = coll.defaultdict(str) # Nested dictionary for tracking all TCR-associated data
                    thimble_dict['linked'] = {'Warnings/Errors': '', 'linkable': True, 'Linker': ''}
                    thimble_dict['out'] = sorted_row_bits

                    # Need to run the stitch function on each individual submitted chain
                    for c in [r1, r2]:
                        chain_notes = ''

                        with warnings.catch_warnings(record=True) as chain_warnings:
                            warnings.simplefilter("always")

                            # Generate the necessary fields for stitching
                            tcr_bits = {'skip_c_checks': input_args['skip_c_checks'], 'species': species,
                                        'mode': input_args['mode']}

                            if 'seamless' in input_args:
                                if input_args['seamless']:
                                    tcr_bits['seamless'] = True

                            # # Convert naming to output formats
                            for field in [x for x in sorted_row_bits if c in x]:
                                if field in convert_fields:
                                    if sorted_row_bits[field]:
                                        tcr_bits[convert_fields[field]] = sorted_row_bits[field]

                            # Determine whether there's supposed to be a TCR for this chain
                            # At the very least we need the relevant TCR info (V/J/CDR3)
                            featured_bits = [x for x in tcr_bits if x in ['v', 'j', 'cdr3']]

                            # If no TCR features present, just skip
                            if len(featured_bits) == 0:
                                thimble_dict['linked']['linkable'] = False
                                continue

                            elif len(featured_bits) in [1, 2]:
                                warnings.warn("Incomplete TCR information - need V/J/CDR3 as minimum.")
                                thimble_dict['linked']['linkable'] = False

                            else:
                                tcr_bits = fxn.autofill_input(populate_blanks(tcr_bits, pre_stitch_list_fields), c)
                                tcr_bits = fxn.tweak_thimble_input(tcr_bits)
                                try:
                                    # Actually run the stitching of this chain
                                    thimble_dict[c] = st.stitch(tcr_bits, tcr_dat[c], tcr_functionality[c], partial,
                                                                codons, input_args['j_warning_threshold'],
                                                                preferences[c], c_res, j_res, low_conf_js)

                                    thimble_dict['out'][c + '_nt'] = thimble_dict[c]['stitched_nt']

                                    # Skip translation of single chains if the 'only_linked' option is selected
                                    if not input_args['only_linked']:
                                        thimble_dict[c]['stitched_aa'] = fxn.pad_trans(thimble_dict[c]['stitched_nt'],
                                                                                       thimble_dict[c]['translation_offset'])
                                        thimble_dict['out'][c + '_aa'] = thimble_dict[c]['stitched_aa']
                                    else:
                                        thimble_dict[c]['stitched_aa'], thimble_dict['out'][c + '_aa'] = '', ''

                                    thimble_dict['out'].update(dict(list(zip(
                                        [c + x for x in post_stitch_list_fields], thimble_dict[c]['out_list']))))

                                except Exception as message:
                                    chain_notes += str(message) + 'Cannot stitch a sequence for ' + c + '. '
                                    thimble_dict['linked']['linkable'] = False

                            # Store all chain related warning messages (ignoring irrelevant errors)
                            # NB only save chain specific data if there's a rearranged chain (for single chain outputs)
                            # but all chain data will be stored in the final all encompassing entry
                            chain_notes += ' '.join(
                                [str(chain_warnings[x].message) for x in range(len(chain_warnings))
                                 if 'DeprecationWarning' not in str(chain_warnings[x].category)])
                            if c in thimble_dict.keys():
                                thimble_dict[c]['Warnings/Errors'] = chain_notes

                        if chain_notes:
                            thimble_dict['linked']['Warnings/Errors'] += chain_notes.replace('.', ' (' + c + ').')

                    with warnings.catch_warnings(record=True) as link_warnings:
                        warnings.simplefilter("always")

                        # If sequences are to be linked, determine the appropriate linker sequence and join together
                        if sorted_row_bits['Linker'] and thimble_dict['linked']['linkable'] == True:

                            if sorted_row_bits['Link_order']:
                                thimble_dict['linked']['Link_order'] = sorted_row_bits['Link_order'].upper()

                                # Only allow valid orders
                                if sorted_row_bits['Link_order'] not in [r2[2]+r1[2], r1[2]+r2[2]]:
                                    warnings.warn("Error: given link order not valid (not " + r1[2]+r2[2] + " or " +
                                                  r2[2]+r1[2] + ") - defaulting to " + r2[2]+r1[2] + ".")
                                    thimble_dict['linked']['Link_order'] = r2[2]+r1[2]

                            else:
                                # Default option is B
                                thimble_dict['linked']['Link_order'] = r2[2]+r1[2]

                            # Extract just the defining locus character (e.g. A/B) to infer proper link order
                            tr1, tr2 = thimble_dict['linked']['Link_order'][0], thimble_dict['linked']['Link_order'][1]

                            if thimble_dict[r1]['stitched_nt'] and thimble_dict[r2]['stitched_nt']:
                                try:
                                    thimble_dict['linked']['Linker_id'] = sorted_row_bits['Linker']
                                    thimble_dict['linked']['linker_nt'] = fxn.get_linker_seq(sorted_row_bits['Linker'],
                                                                                             linker_dict)

                                    thimble_dict['linked']['Linked_nt'] = thimble_dict['TR' + tr1]['stitched_nt'] + \
                                                                          thimble_dict['linked']['linker_nt'] + \
                                                                          thimble_dict['TR' + tr2]['stitched_nt']
                                    thimble_dict['linked']['Linked_aa'] = fxn.pad_trans(thimble_dict['linked']['Linked_nt'],
                                                                                        thimble_dict[r1]
                                                                                        ['translation_offset'])

                                    # Add warnings if sequences applied at maybe the wrong ends of things
                                    if (thimble_dict[r1]['in']['5_prime_seq']
                                        and thimble_dict['linked']['Link_order'] == r2[2]+r1[2]) or \
                                            (thimble_dict[r2]['in']['5_prime_seq']
                                             and thimble_dict['linked']['Link_order'] == r1[2]+r2[2]):
                                        warnings.warn("Warning: " + thimble_dict['linked']['Link_order'] + " order "
                                                    "specified, but 3' chain has an additional 5' sequence provided. ")

                                    if (thimble_dict[r1]['in']['3_prime_seq']
                                        and thimble_dict['linked']['Link_order'] == r1[2]+r2[2]) or \
                                            (thimble_dict[r2]['in']['3_prime_seq']
                                             and thimble_dict['linked']['Link_order'] == r2[2]+r1[2]):
                                        warnings.warn("Warning: " + thimble_dict['linked']['Link_order'] + " order "
                                                    "specified, but 5' chain has an additional 3' sequence provided. ")

                                # Store any error messages
                                except Exception as message:
                                    thimble_dict['linked']['Warnings/Errors'] += str(message)

                            else:
                                warnings.warn("Error: need both a " + r1 + " and " + r2 + " to link. ")

                        # Add any warnings/errors derived from the linkage, or either constituent chain
                        if thimble_dict['linked']['Warnings/Errors'] or link_warnings:
                            thimble_dict['linked']['Warnings/Errors'] += ' '.join(
                                [str(link_warnings[x].message) for x in range(len(link_warnings))
                                 if 'DeprecationWarning' not in str(link_warnings[x].category)])

                        if not thimble_dict['linked']['Warnings/Errors']:
                            thimble_dict['linked']['Warnings/Errors'] = "[None]"

                        # Move over the remaining fields from the 'linked' dict to the 'out' dict
                        for f in ['Linked_nt', 'Linked_aa']:
                            if f in thimble_dict['linked']:
                                thimble_dict['out'][f] = thimble_dict['linked'][f]
                            else:
                                thimble_dict['out'][f] = ''

                    thimble_dict['out']['Warnings/Errors'] = thimble_dict['linked']['Warnings/Errors']

                    # Store output as one long string, for a single write-out once input file is finished
                    for field in [x for x in out_headers[receptor] if x not in thimble_dict['out']]:
                        thimble_dict['out'][field] = thimble_dict['linked'][field]

                    out_data.append('\t'.join([thimble_dict['out'][x] for x in out_headers[receptor]]))

                    # If requested, also output additional more detailed file types
                    if input_args['mode'] == 'JSON':
                        thimble_dict['metadata'] = fxn.get_metadata_dict(input_args['species'], 'thimble', __version__)
                        with open(os.path.join(input_args['out_folder'], thimble_dict['out']['TCR_name'])
                                  + '.json', 'w') as out_file:
                            json.dump(thimble_dict, out_file)

                    elif input_args['mode'] == 'GB':

                        # Generate files for each chain, unless linked only option selected
                        subdirs = ['linked']
                        if not input_args['only_linked']:
                            subdirs += [r1, r2]
                        for subdir in subdirs:
                            if subdir not in thimble_dict:
                                continue
                            spec_description = base_description
                            if thimble_dict[subdir]['Warnings/Errors'] != "[None]":
                                spec_description +=  ("\n" + fxn.spacer + "\n" + fxn.spacer +
                                                      "This rearrangement produced the following warning messages:"
                                                      "\n" + fxn.spacer + thimble_dict[subdir]['Warnings/Errors'])

                            # TODO move stitchr/thimble-to-genbank formatting to own function?
                            gb_feats = []
                            if subdir != 'linked':
                                gb_feats = []
                                for feat in thimble_dict[subdir]['indexes']:
                                    gb_feats.append((thimble_dict[subdir]['used'][feat], '',
                                                     [thimble_dict[subdir]['seqs'][feat]],
                                                     [thimble_dict[subdir]['indexes'][feat]]))
                            else:
                                if not thimble_dict[subdir]['linkable']:
                                    continue
                                # Establish the order and re-engineer up the indexes to ensure correct linked placement
                                thimble_dict[subdir]['stitched_nt'] = thimble_dict[subdir]['Linked_nt']

                                sorted_chain1 = sorted(thimble_dict['TR' + tr1]['indexes'].items(), key=lambda x: x[1])
                                sorted_chain2 = sorted(thimble_dict['TR' + tr2]['indexes'].items(), key=lambda x: x[1])

                                for feat in sorted_chain1:
                                    gb_feats.append((thimble_dict['TR' + tr1]['used'][feat[0]], '',
                                                    [thimble_dict['TR' + tr1]['seqs'][feat[0]]],
                                                    [thimble_dict['TR' + tr1]['indexes'][feat[0]]]))

                                offset = len(thimble_dict['TR' + tr1]['stitched_nt'])
                                gb_feats.append((thimble_dict['linked']['Linker_id'], '',
                                                    [thimble_dict['linked']['linker_nt']],
                                                    [offset]))
                                offset += len(thimble_dict['linked']['linker_nt'])

                                for feat in sorted_chain2:
                                    gb_feats.append((thimble_dict['TR' + tr2]['used'][feat[0]], '',
                                                    [thimble_dict['TR' + tr2]['seqs'][feat[0]]],
                                                    [thimble_dict['TR' + tr2]['indexes'][feat[0]] + offset]))

                            genbank_params = {'sequence_name': thimble_dict['out']['TCR_name'],
                                              'full_sequence': thimble_dict[subdir]['stitched_nt'],
                                              'description': spec_description, 'topology': 'linear',
                                              'features': gb_feats,
                                              'save_dir_path': os.path.join(input_args['out_folder'], subdir),
                                              'species': input_args['species'],
                                              'numbering': False,
                                              'plot_multi': True, 'division': 'SYN', 'journal': 'thimble'}

                            fxn.output_genbank(**genbank_params)

                tcr_count += 1

            line_count += 1

    # TODO delete unused output folders for GB/JSON output?

    time_taken = time() - start
    print("Took", str(round(time_taken, 2)), "seconds")

    # Write out data
    with open(os.path.join(input_args['out_folder'], input_args['out_file']), 'w') as out_file:
        out_string = '\n'.join(out_data)
        out_file.write(out_string)



# -*- coding: utf-8 -*-

"""
stitchr.py

Takes command line V, J and (amino acid) CDR3 information
to generate a full length coding nucleotide TCR sequence.
Can be used for TCR vector design, and other purposes.

"""

from . import stitchrfunctions as fxn
import argparse
import sys
import warnings


__version__ = '1.2.0'
__author__ = 'Jamie Heather'
__email__ = 'jheather@mgh.harvard.edu'

sys.tracebacklimit = 0  # comment when debugging
warnings.formatwarning = fxn.custom_formatwarning

def args():
    """
    args(): Obtains command line arguments which dictate the script's behaviour
    """

    # Help flag
    parser = argparse.ArgumentParser(
        description="stiTChR v" + str(__version__) + '\n' +
                    ": Stitch together a coding TCR nucleotide sequence from V, J, and CDR3 info.\n"
                    "Use IMGT gene names, and include terminal CDR3 residues (C/F).\n\n"
                    "E.g. \'stitchr -v TRBV20-1*01 -j TRBJ1-2*01 -cdr3 CASWHATEVERF\'.\n"
                    "See https://github.com/JamieHeather/stitchr and https://doi.org/10.1093/nar/gkac190.")

    # Input and output options
    parser.add_argument('-v', '--v', required=True, type=str, default='', help="V gene name. Required. "
                        "Specific allele not required, will default to prototypical (*01)")

    parser.add_argument('-j', '--j', required=True, type=str, default='', help="J gene name. Required. "
                        "Specific allele not required, will default to prototypical (*01)")

    parser.add_argument('-cdr3', '--cdr3', required=True, type=str, default='',
                        help="CDR3 amino acid sequence. Required. Must include terminal residues (e.g. C/F)")

    parser.add_argument('-s', '--species', required=False, type=str, default='HUMAN',
                        help="Species (common name). Optional: see data directory for all possible options. "
                             "Default = HUMAN")

    parser.add_argument('-c', '--c', required=False, type=str, default='', help="Constant gene. Optional. "
                        "Specific allele not required, will default to prototypical (*01). "
                        "See README re: alternative TRGC exon configurations.")

    parser.add_argument('-l', '--l', required=False, type=str, default='', help="Leader region. Optional. "
                        "Will default to match the appropriate V gene.")

    parser.add_argument('-aa', '--aa', required=False, type=str, default='',
                        help="Partial amino acid sequence, if known. Optional. Can be used to check stitching success.")

    parser.add_argument('-n', '--name', required=False, type=str, default='',
                        help="Name for TCR sequence. Optional. Will be added to output FASTA header.")

    parser.add_argument('-sl', '--seamless', action='store_true', required=False, default=False,
                        help="Optional flag to integrate known nucleotide sequences seamlessly. \n NB: "
                        "nucleotide sequences covering the CDR3 junction with additional V gene context required.")

    parser.add_argument('-5p', '--5_prime_seq', required=False, type=str, default='',
                        help="Optional sequence to add to the 5' of the output sequence (e.g. a Kozak sequence).")

    parser.add_argument('-3p', '--3_prime_seq', required=False, type=str, default='',
                        help="Optional sequence to add to the 3' out the output sequence (e.g. a stop codon).")

    parser.add_argument('-xg', '--extra_genes', action='store_true', required=False, default=False,
                        help="Optional flag to use additional (non-database/-natural) sequences in the "
                             "'additional-genes.fasta' file.")

    parser.add_argument('-p', '--preferred_alleles_path', required=False, type=str, default='',
                        help="Path to a file of preferred alleles to use when no allele specified (instead of *01). "
                             "Optional.")

    parser.add_argument('-m', '--mode', required=False, type=str, default='both_fa',
                        help="Standard out output mode. "
                             "Options are 'BOTH_FA' (default), 'AA_FA', 'NT_FA', 'AA', 'NT', 'GB', 'JSON'.")

    parser.add_argument('-cu', '--codon_usage_path', required=False, type=str, default='',
                        help="Path to a file of Kazusa-formatted codon usage frequencies. Optional.")

    parser.add_argument('-jt', '--j_warning_threshold', required=False, type=int, default=3,
                        help="J gene substring length warning threshold. Default = 3. "
                             "Decrease to get fewer notes on short J matches.")

    parser.add_argument('-sc', '--skip_c_checks', action='store_true', required=False, default=False,
                        help="Optional flag to skip usual constant region gene checks.")

    parser.add_argument('-sw', '--suppress_warnings', action='store_true', required=False, default=False,
                        help="Optional flag to suppress warnings.")

    parser.add_argument('--version', action='version', version=__version__,
                        help="Print current stitchr version.")

    parser.add_argument('--cite', action=fxn.GetCitation, help="Print citation details.", nargs=0)

    parser.add_argument('-dd', '--data_dir', action='version', version=fxn.data_dir,
                        help="Print installed stitchr data directory path.")

    return parser.parse_args()


def stitch(specific_args, tcr_info, functionality, partial_info, codon_dict, j_warning_threshold, preferences,
           c_motifs, j_residues, low_confidence_js):
    """
    Core function, that performs the actual TCR stitching
    :param specific_args: dict, basic input arguments of a given rearrangement (e.g. V/J/CDR3)
    :param tcr_info: dict, germline sequence data for the alleles of a specific locus
    :param functionality: dict, predicted functionality of different TCR genes
    :param partial_info: dict, genes filtered out from input TCR data on account of being incomplete in the database
    :param codon_dict: dictionary of which codons to use for which amino acids
    :param j_warning_threshold: int threshold value, if a J substring length match is shorter it will throw a warning
    :param preferences: nested dict of preferred alleles, like the tcr_info dict but one level shallower
    :param c_motifs: dict, product of get_c_motifs, detailing N terminal C gene translations to find correct ORFs
    :param j_residues: dict, product of get_j_motifs, detailing the conserved F/equivalent residue to look for
    :param low_confidence_js: list of str detailing 'low confidence' J regions which lack an obvious FGXG motif
    :return: stitch_dict of details of the in/in process/constructed TCR,
        including the stitched nt and translation offset (int, 0/1/2)
    """

    stitch_dict = {'in': specific_args, 'used': {}, 'seqs': {}}

    for r in fxn.regions:

        # First establish what the input gene and allele values are
        gene_allele_in = specific_args[r]
        if '*' not in gene_allele_in:
            gene_allele_in += '*'

        in_gene, in_allele = gene_allele_in.split('*')
        gene, allele = '', ''

        # First check whether the gene exists in the input data for this species
        if in_gene in tcr_info[fxn.regions[r]]:
            gene = in_gene

        else:
            # If it's a leader sequence, it might be a user-defined DNA sequence
            if r == 'l' and fxn.dna_check(specific_args['l']):
                # If it is, add that info in...
                stitch_dict['used'][r] = 'UserSpecifiedLeader*' + specific_args['l']
                stitch_dict['seqs'][r] = specific_args['l']

                # Check it's likely to translate in frame
                if len(specific_args['l']) % 3 != 0:
                    warnings.warn("User specified leader sequence is not evenly divisible by 3 - "
                                  "stitched TCR frame will likely be wrong. ")

                # ...and jump ahead to the stitching (skipping irrelevant TCR gene checks)
                continue

            raise ValueError("Error: a " + fxn.regions[r] + " sequence region has not been found for gene " + gene +
                             " in the IMGT data for this chain/species. Please check your TCR and species data. ")

        # Having established the gene, then need to determine the allele
        if in_allele:

            # If an allele is provided, check whether it exists in the database for this gene and isn't partial
            if in_allele in tcr_info[fxn.regions[r]][gene]:

                if partial_info[gene][allele]:
                    warnings.warn("Cannot use " + gene + "*" + in_allele + " " + fxn.regions[r].lower() + "sequence, "
                                  "as it is classed as '" + partial_info[gene][in_allele] + "' (partial). ")
                else:
                    allele = in_allele

            else:
                warnings.warn("Cannot find the sequence of the requested allele, " + gene + "*" + in_allele +
                              ", for the " + fxn.regions[r].lower() + " sequence in the input FASTA data. ")

        # If no allele supplied (or is supplied but invalid) then use 1) a preferred allele or 2) the prototypical *01
        if not allele:

            warnings.warn("No valid " + fxn.regions[r].lower() + " region allele determined yet for " + gene + ". ")
            allele = '01'

            if preferences:
                if gene in preferences[fxn.regions[r]]:
                    allele = preferences[fxn.regions[r]][gene]
                    warnings.warn("Defaulting to allele *" + allele + " for the " + fxn.regions[r].lower() +
                                  " sequence, as specified in the preferred allele file. ")
                    # NB: we don't have to worry about partial alleles in the preference list, as those are filtered out
                else:
                    warnings.warn("Defaulting to *01, as " + gene + " isn't specified in the preferred allele file for"
                                                                    " the " + fxn.regions[r].lower() + " region. ")
            else:
                warnings.warn("Defaulting to *01 for the " + fxn.regions[r].lower() + " region, "
                              "in the absence of a preferred allele file being specified. ")

            # Check how feasible the defaulted allele is
            if allele == '01':
                # First check this allele actually exists - if not pick one from what's available
                if not tcr_info[fxn.regions[r]][gene][allele]:
                    new_allele = [x for x in tcr_info[fxn.regions[r]][gene] if tcr_info[fxn.regions[r]][gene][x]][0]
                    warnings.warn("No sequence found for " + gene + "*" + allele + " for the " + fxn.regions[r] +
                                  " region. " + gene + "*" + new_allele + " is being used instead - please double "
                                  "check either the intended allele or that this is a suitable replacement. ")
                    allele = new_allele

                # If the prototypical allele does exist, flag a warning if there are other options available
                elif len(tcr_info[fxn.regions[r]][gene]) > 1:
                    # Just check that at least one of those other alleles is not partial
                    for other_allele in tcr_info[fxn.regions[r]][gene]:
                        if other_allele != '01':
                            if not partial_info[gene][other_allele]:
                                warnings.warn("NB: the prototypical '*01' allele is being used for the " +
                                              fxn.regions[r].lower() + " region by default, but other alleles are "
                                                                       "available - consider double checking the right allele is asked for. ")
                                continue

        # Catchall double check both gene and allele are sorted
        if gene and allele:
            stitch_dict['seqs'][r] = tcr_info[fxn.regions[r]][gene][allele]
            stitch_dict['used'][r] = gene + '*' + allele

            # Add '(L)' to leader names
            if r == 'l':
                stitch_dict['used']['l'] = stitch_dict['used']['l'] + '(L)'

            func_err_base = "Warning: gene " + gene + '*' + allele + " has an assigned functionality of \'" \
                            + functionality[gene][allele] + "\', "

            # Check functionality
            if fxn.strip_functionality(functionality[gene][allele]) != 'F':
                warnings.warn(func_err_base + "and thus may not express or function correctly. ")

            # Special check to account for IMGT not including the 3' terminal J residue if allele from cDNA
            if functionality[gene][allele] == '(F)' and r == 'j':

                cdna_j_err = func_err_base + "meaning it was only detected in cDNA, and thus IMGT doesn't record its" \
                                             " 3\' terminal nucleotide - "

                # If this is an (F) call for a prototypical allele we have to skip, as there's no reference to go off
                if allele == '01':
                    raise IOError(cdna_j_err + " unable to fix as this is the prototypical allele (*01). ")
                # Otherwise, use the 01 terminal residue
                else:
                    stitch_dict['seqs'][r] = stitch_dict['seqs'][r] + tcr_info[fxn.regions[r]][gene]['01'][-1]
                    warnings.warn(cdna_j_err + "substituting the *01 allele terminal base to maintain reading frame. ")

        else:
            raise ValueError(
                "Cannot find TCR sequence data for " + r.upper() + " gene: " + in_gene + '*' + in_allele + ". ")

    # Throw a warning if the J gene is one in which the C-terminal residue cannot be confidently identified
    if stitch_dict['used']['j'] in low_confidence_js:
        warnings.warn("Warning: " + stitch_dict['used']['j'] + " has a \'low confidence\' CDR3-ending motif. ")

    # Then determine whether CDR3 has been provided in amino or nucleic acid form
    if fxn.dna_check(specific_args['cdr3']):
        stitch_dict['input_type'] = 'nt'
        stitch_dict['used']['cdr3'] = fxn.translate_nt(specific_args['cdr3'])
        warnings.warn("CDR3 junction provided as DNA sequence: \'" + specific_args['cdr3'] + '\'. ')
    else:
        stitch_dict['input_type'] = 'aa'
        stitch_dict['used']['cdr3'] = specific_args['cdr3']

    # Determine whether seamless integrated is requested (in platform independent way)
    seamless = False
    if 'seamless' in specific_args:
        if specific_args['seamless']:
            seamless = True
    else:
        specific_args['seamless'] = False
        stitch_dict['in']['seamless'] = False

    # Get the germline encoded bits
    n_term_nt_raw = stitch_dict['seqs']['l'] + stitch_dict['seqs']['v']
    c_term_nt_raw = stitch_dict['seqs']['j'] + stitch_dict['seqs']['c']

    # Run the appropriate form of non-templated integration
    # First test eligibility for seamless integration
    if stitch_dict['input_type'] == 'nt' and seamless:

        warnings.warn(
            "Seamless option selected: stitched sequence may not be accurate if provided NT sequence is too short"
            " or contains polymorphisms/errors near the recombining edges. ")

        # Optimistic warning to ensure additional sequence provided for seamless stitching
        if fxn.translate_nt(specific_args['cdr3'][:3]) == 'C':
            warnings.warn("Cys residue detected in first codon of provided seamless CDR3 junction: "
                          "note that seamless stitching requires additional sequence 5\' of the start of the CDR3 "
                          "(disregard this warning if this is not the CDR3 starting residue). ")

        n_term_nt_trimmed, v_overlap = fxn.find_v_overlap(n_term_nt_raw, specific_args['cdr3'])

        # Check for suspiciously short V gene overlaps
        if len(v_overlap) < 10:
            warnings.warn("Only short V gene overlap detected (" + v_overlap + ") for seamless stitching. ")

            # Most common cause = unexpected polymorphism (e.g. SNP or PCR error) in the 5' of the padding sequence
            # Try the overlap search again, starting from one position upstream of the previous match
            n_term_nt_trimmed_2, v_overlap_2 = fxn.find_v_overlap(n_term_nt_raw,
                                                                  specific_args['cdr3'][len(v_overlap)+1:])

            if len(v_overlap_2) > 10:
                warnings.warn("A longer (" + str(len(v_overlap_2)) + " nt) overlap was found after trimming the "
                                                                     "short overlap +1 off, and stitching continued (NB: presumed SNP or PCR error). ")

                n_term_nt_trimmed = n_term_nt_trimmed_2[:-(len(v_overlap) + 1)]
                v_overlap = specific_args['cdr3'][:len(v_overlap) + len(v_overlap_2) + 1]

            else:
                raise ValueError("No longer overlap was found even after trimming that short overlap + 1. Please check "
                                 "the V gene call and that the CDR3 padding sequence doesn't contain polymorphisms. ")

        c_term_nt_trimmed = fxn.find_j_overlap(specific_args['cdr3'][len(v_overlap):], c_term_nt_raw)
        stitched_nt = n_term_nt_trimmed + specific_args['cdr3'] + c_term_nt_trimmed

        stitch_dict['seqs']['seamless_insert'] = specific_args['cdr3']
        stitch_dict['used']['seamless_insert'] = 'seamless_insert'

        # Use the tidy_c_term functionality to frame check/trim excess
        stitched_nt, stitched_trans = fxn.tidy_c_term(stitched_nt, False, c_motifs, stitch_dict['used']['c'])

        # Catch more 5' SNP errors: if there's a SNP in the edges of the contextual padding this can cause and indel,
        # ... which means tidy_c_term will trim some nt from the 5' of the gene, which we can use to trigger an IOError
        if not stitched_nt.startswith(stitch_dict['seqs']['l']):
            raise ValueError("An indel has been detected during seamless stitching, which is usually caused by "
                             "polymorphisms in the padding sequence relative to the genes selected: please either "
                             "ensure selected alleles are correct or provide more context beyond any polymorphisms. ")

    # Otherwise run regular amino-acid based germline determination
    else:

        # If an exact nucleotide junction is provided, work off the translated product
        if stitch_dict['input_type'] == 'nt':

            # Frame check (if not providing extra context for seamless integration)
            if len(stitch_dict['used']['cdr3']) % 3 != 0 and not seamless:
                warnings.warn("Warning: length of CDR3 DNA sequence provided is not evenly divisible by 3 "
                              "and seamless stitching not selected: stitched TCR frame will likely be wrong. ")

        # And check that users haven't asked for amino acid/seamless, which won't work
        elif seamless:
            raise IOError("The seamless option has been selected, yet provided CDR3-containing sequence is not DNA. ")

        # Get codon data, and use to check that there's no unexpected characters in the CDR3
        if len([x for x in list(set([x for x in stitch_dict['used']['cdr3']])) if x not in list(codon_dict.keys())]) > 0:
            raise ValueError("Unexpected character in CDR3 string. "
                             "Please use only one-letter standard amino acid designations. ")

        # Then check the C-terminus of the CDR3 has an appropriate residue
        # (putting the default F in the dict if not there)
        if stitch_dict['used']['j'] not in j_residues:
            j_residues[stitch_dict['used']['j']] = 'F'

        if stitch_dict['used']['cdr3'][-1] != j_residues[stitch_dict['used']['j']]:
            warnings.warn("CDR3 provided does not end with the expected residue for this J gene (" +
                          j_residues[stitch_dict['used']['j']] + "). Deletion this far in to the J is extremely unlikely. ")

        # Tidy up the germline edges to be coding for whole codons without any remainders
        n_term_nt_inframe, n_term_aa = fxn.tidy_n_term(n_term_nt_raw)

        c_term_nt_inframe, c_term_aa = fxn.tidy_c_term(c_term_nt_raw,
                                                       specific_args['skip_c_checks'],
                                                       c_motifs, stitch_dict['used']['c'])

        # Figure out where the AA CDR3 will slot in: look at the CDR3 edges & see how much overlap needs to be removed
        # Start with 4 residues chunks, move from end of V gene up to 10 residues in (very generous deletion allowance)
        n_term_nt_trimmed, cdr3_n_offset = fxn.determine_v_interface(stitch_dict['used']['cdr3'],
                                                                     n_term_nt_inframe, n_term_aa)
        c_term_nt_trimmed, cdr3_c_end = fxn.determine_j_interface(stitch_dict['used']['cdr3'][cdr3_n_offset:],
                                                                  c_term_nt_inframe, c_term_aa,
                                                                  len(stitch_dict['seqs']['j']), j_warning_threshold)

        # Generate the non-templated sequences using either supplied nucleotides or common codons established earlier
        if stitch_dict['input_type'] == 'nt':
            non_templated_nt = specific_args['cdr3'][cdr3_n_offset * 3:(cdr3_n_offset+cdr3_c_end) * 3]
        else:
            non_templated_aa = stitch_dict['used']['cdr3'][cdr3_n_offset:cdr3_n_offset+cdr3_c_end]
            non_templated_nt = fxn.rev_translate(non_templated_aa, codon_dict)

        # Then finally stitch all that info together and output!
        stitched_nt = n_term_nt_trimmed + non_templated_nt + c_term_nt_trimmed
        stitch_dict['seqs']['non_templated'] = non_templated_nt
        stitch_dict['used']['non_templated'] = 'non_templated'

    stitch_dict['seqs']['v'] = n_term_nt_trimmed[len(stitch_dict['seqs']['l']):]
    stitch_dict['seqs']['c'] = stitch_dict['seqs']['c'][:stitch_dict['seqs']['c'].index(c_term_nt_trimmed[-20:]) + 20]
    stitch_dict['seqs']['j'] = c_term_nt_trimmed[:-len(stitch_dict['seqs']['c'])]
    # TODO potentially find a more elegant way to determine/retain the C start/end positions?

    # If optional 5'/3' sequences are specified, add them to the relevant place
    if specific_args['5_prime_seq']:

        stitched_nt = specific_args['5_prime_seq'] + stitched_nt
        stitch_dict['seqs']['5_prime_seq'] = specific_args['5_prime_seq']
        stitch_dict['used']['5_prime_seq'] = '5_prime_seq'

        # Translation offset allows simple translation of output NT without having to figure out the frame
        stitch_dict['translation_offset'] = 3 - (len(specific_args['5_prime_seq']) % 3)
        if stitch_dict['translation_offset'] == 3:
            stitch_dict['translation_offset'] = 0

    else:
        stitch_dict['translation_offset'] = 0

    if specific_args['3_prime_seq']:
        stitch_dict['used']['3_prime_seq'] = '3_prime_seq'
        stitched_nt += specific_args['3_prime_seq']
        stitch_dict['seqs']['3_prime_seq'] = specific_args['3_prime_seq']

    stitch_dict['out_list'] = [specific_args['name'], stitch_dict['used']['v'], stitch_dict['used']['j'],
                               stitch_dict['used']['c'], specific_args['cdr3'], stitch_dict['used']['l']]
    stitch_dict['stitched_nt'] = stitched_nt

    # If this is to be output to GenBank format, also determine the location of each substr,
    # to ensure proper plotting of even non-unique string
    if specific_args['mode'] == 'GB':
        stitch_dict['indexes'] = {}
        region_order = ['5_prime_seq', 'l', 'v', 'cdr3', 'seamless_insert', 'j', 'c', '3_prime_seq']

        # Generate the nt sequence of the CDR3 (where possible)
        if not stitch_dict['in']['seamless']:
            translation = fxn.translate_nt('N' * stitch_dict['translation_offset'] + stitch_dict['stitched_nt'])
            cdr3_match = translation.find(stitch_dict['used']['cdr3']) * 3
            stitch_dict['seqs']['cdr3'] = stitch_dict['stitched_nt'][cdr3_match - stitch_dict['translation_offset']:
                                                                     cdr3_match + stitch_dict['translation_offset'] + len(stitch_dict['used']['cdr3']*3)]

        loc = 0
        for region in region_order:
            if region in stitch_dict['seqs'].keys():
                if region == '3_prime_seq':
                    stitch_dict['indexes'][region] = stitched_nt.rfind(stitch_dict['seqs'][region])
                else:
                    index = stitched_nt[loc:].index(stitch_dict['seqs'][region])
                    loc += index
                    stitch_dict['indexes'][region] = loc

        if 'non_templated' in stitch_dict['seqs'].keys():
            stitch_dict['indexes']['non_templated'] = (stitched_nt[:stitch_dict['indexes']['j']].rfind(
                stitch_dict['seqs']['non_templated']))

    return stitch_dict


gene_types = list(fxn.regions.values())


def main():

    # Get input arguments, determine the TCR chain in use, get codon table, then load the IMGT data in
    input_args, chain = fxn.sort_input(vars(args()))
    codons = fxn.get_optimal_codons(input_args['codon_usage_path'], input_args['species'])
    imgt_dat, tcr_functionality, partial = fxn.get_ref_data(chain, gene_types, input_args['species'])

    # Get information about the C-terminal residue of the CDR3 *should* be, given that J gene
    j_res, low_conf_js = fxn.get_j_motifs(input_args['species'])
    # And the motifs required for the correct frame inference and delineation of the constant region sequences
    c_res = fxn.get_c_motifs(input_args['species'])

    if input_args['suppress_warnings']:
        warnings.filterwarnings('ignore')

    if input_args['extra_genes']:
        imgt_dat, tcr_functionality = fxn.get_additional_genes(imgt_dat, tcr_functionality)
        input_args['skip_c_checks'] = True

    if input_args['preferred_alleles_path']:
        preferred_alleles = fxn.get_preferred_alleles(input_args['preferred_alleles_path'],
                                                      gene_types, imgt_dat, partial, chain)
    else:
        preferred_alleles = {}

    with warnings.catch_warnings(record=True) as stitch_warnings:
        stitched = stitch(input_args, imgt_dat, tcr_functionality, partial, codons, input_args['j_warning_threshold'],
                          preferred_alleles, c_res, j_res, low_conf_js)

    out_str = '|'.join(stitched['out_list'])

    # Output the appropriate strings to stdout
    if input_args['mode'] not in ['BOTH_FA', 'AA_FA', 'NT_FA', 'AA', 'NT', 'GB', 'JSON']:
        raise IOError("Unknown output mode detected: " + input_args['mode'] + ". \n"
                      "Should be one of 'BOTH_FA' (default), 'AA_FA', 'NT_FA', 'AA', 'NT', 'GB', or 'JSON'.")

    if '_FA' in input_args['mode']:
        print('----------------------------------------------------------------------------------------------')
        if input_args['mode'] == 'BOTH_FA' or input_args['mode'] == 'NT_FA':
            print(fxn.fastafy('nt|' + out_str, stitched['stitched_nt']))

        if input_args['mode'] == 'BOTH_FA' or input_args['mode'] == 'AA_FA':
            # Use the offset to 5' pad the stitched sequence with 'N's to make up for non-codon length 5' added seqs
            print(fxn.fastafy('aa|' + out_str,
                              fxn.translate_nt('N' * stitched['translation_offset']
                                               + stitched['stitched_nt'])))

    elif input_args['mode'] == 'NT':
        print(stitched['stitched_nt'])

    elif input_args['mode'] == 'AA':
        print(fxn.translate_nt('N' * stitched['translation_offset'] + stitched['stitched_nt']))

    elif input_args['mode'] == 'GB':

        gb_file_name = fxn.get_output_name(stitched)

        # Generate the DEFINITION field
        description = fxn.get_metadata_text(input_args['species'], 'stitchr', __version__)

        if input_args['extra_genes']:
            description += ("Note that this was run using the 'extra genes' flag, and thus additional non-reference "
                            "sequences may have been incorporated. ")

        description += ("\n" + fxn.spacer + "\n" + fxn.spacer +
                        "Produced using the stitchr command:\n" + fxn.spacer + "'" + ' '.join(sys.argv[:]) + "'. ")

        if stitch_warnings and not input_args['suppress_warnings']:
            description += ("\n" + fxn.spacer + "\n" + fxn.spacer +
                            "This rearrangement produced the following warning messages:\n" + fxn.spacer) + ' '.join(
                [str(stitch_warnings[x].message) for x in range(len(stitch_warnings))])

        # And define the individual features to be recorded
        gb_feats = []
        for feat in stitched['indexes']:
            gb_feats.append((stitched['used'][feat], '', [stitched['seqs'][feat]], [stitched['indexes'][feat]]))
        # TODO move stitchr/thimble-to-genbank formatting to own function?

        genbank_params = {'sequence_name': gb_file_name, 'full_sequence': stitched['stitched_nt'],
                          'description': description, 'topology': 'linear', 'features': gb_feats,
                          'save_dir_path': './', 'species': input_args['species'], 'numbering': False,
                          'plot_multi': True, 'division': 'SYN', 'journal': 'stitchr'}

        fxn.output_genbank(**genbank_params)

    elif input_args['mode'] == 'JSON':
        import json
        stitched['metadata'] = fxn.get_metadata_dict(input_args['species'], 'stitchr', __version__)
        stitched['Warnings/Errors'] = ''.join([str(stitch_warnings[x].message)
                                                               for x in range(len(stitch_warnings))])

        with open(fxn.get_output_name(stitched) + '.json', 'w') as out_file:
            json.dump(stitched, out_file)

    # If a known/partial amino acid sequence provided, ensure they match up with a quick printed alignment
    if input_args['aa']:
        from Bio import pairwise2
        from Bio.pairwise2 import format_alignment
        alignments = pairwise2.align.globalxx(input_args['aa'],
                                              fxn.translate_nt('N' * stitched['translation_offset']
                                                               + stitched['stitched_nt']))
        for i in range(0, 600, 60):
            print('\n')
            if i > len(alignments[0][0]):
                break
            for y in [x[i:i+60] for x in format_alignment(*alignments[0]).split('\n')[:3]]:
                print(y)

    # Finally print any warnings generated during the running
    if stitch_warnings:
        print('\n'.join([str(stitch_warnings[x].message) for x in range(len(stitch_warnings))]))




import sys
import os
import pytest
import subprocess
import collections as coll
from Stitchr import stitchrfunctions as fxn


def empty_file(file_name):
    with open(file_name, 'w') as out_file:
        out_file.write('')
    return file_name


def test_downloaded_data_handling():
    # Need to have data downloaded to test various functions; arbitrarily use canine and mouse
    dl_cmd = 'stitchrdl -s dog'
    subprocess.call(dl_cmd, shell=True)
    dl_cmd = 'stitchrdl -s mouse'
    subprocess.call(dl_cmd, shell=True)

    assert 'DOG' in fxn.find_species_covered()
    assert 'MOUSE' in fxn.find_species_covered()

    tmp_file = empty_file('temp_dog.txt')
    assert fxn.infer_species(tmp_file) == 'DOG'
    os.remove(tmp_file)

    tmp_file = empty_file('temp_dog_mouse.txt')
    assert fxn.infer_species(tmp_file) == ''
    os.remove(tmp_file)


def test_file_handling():
    in_header, in_seq = ['test_header', 'ACGT']
    tmp_file = 'temp.fasta'
    with open(tmp_file, 'w') as out_file:
        out_file.write(fxn.fastafy(in_header, in_seq))

    with fxn.opener(tmp_file) as in_file:
        for header, seq, null in fxn.read_fa(in_file):
            assert in_header == header and in_seq == seq, "Failed FASTA handling."

    os.remove(tmp_file)


def test_get_chain():
    assert fxn.get_chain('TRAV', 'TRAJ') == 'TRA'
    assert fxn.get_chain('TRDV', 'TRAJ') == 'TRA'
    assert fxn.get_chain('TRBV', 'TRBJ') == 'TRB'
    assert fxn.get_chain('TRDV', 'TRDJ') == 'TRD'
    assert fxn.get_chain('TRGV', 'TRGJ') == 'TRG'
    assert fxn.get_chain('IGHV', 'IGHJ') == 'IGH'
    assert fxn.get_chain('IGLV', 'IGLJ') == 'IGL'
    assert fxn.get_chain('IGKV', 'IGKJ') == 'IGK'
    assert not fxn.get_chain('TRAV', 'TRAJ') == 'TRB'
    with pytest.raises(ValueError):
        fxn.get_chain('TRAV', 'TRBJ')


def test_get_imgt_data():
    assert fxn.get_ref_data('TRA', list(fxn.regions.values()), 'DOG')

    with pytest.raises(ValueError):
        fxn.get_ref_data('TEST', list(fxn.regions.values()), 'DOG')

    with pytest.raises(IOError):
        fxn.get_ref_data('TRA', list(fxn.regions.values()), 'UNICORN')

    imgt_dat = fxn.get_ref_data('TRA', list(fxn.regions.values()), 'MOUSE')
    assert [isinstance(x, dict) for x in imgt_dat]
    assert list(imgt_dat[0].keys()) == ['LEADER', 'VARIABLE', 'JOINING', 'CONSTANT']
    assert list(imgt_dat[1].keys())[0].startswith('TR')

    preferred = fxn.get_preferred_alleles('templates/preferred-alleles/mouse_c57-bl6_example.tsv',
                                          list(fxn.regions.values()), imgt_dat[0], imgt_dat[2], 'TRA')
    assert preferred['LEADER']['TRAV15D-3'] == '02'


aa = ['*', 'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
def test_translation_handling():
    codon_file = os.path.join(fxn.data_dir, 'kazusa', 'DOG.txt')
    assert fxn.get_codon_frequencies(codon_file)
    codon_freqs = fxn.get_codon_frequencies(codon_file)
    assert [x for x in aa if x in codon_freqs.keys()] == aa
    assert [x for x in aa if x in list(set(fxn.codons.values()))] == aa
    assert fxn.get_optimal_codons(codon_file, 'DOG')
    codons = fxn.get_optimal_codons(codon_file, 'DOG')

    assert (fxn.translate_nt('AAAAACAAGAATACAACCACGACTAGAAGCAGGAGTATAATCATGATTCAACACCAGCATCCAC'
                             'CCCCGCCTCGACGCCGGCGTCTACTCCTGCTTGAAGACGAGGATGCAGCCGCGGCTGGAGGCGG'
                             'GGGTGTAGTCGTGGTTTAATACTAGTATTCATCCTCGTCTTGATGCTGGTGTTTATTCTTGTTT') ==
            'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSS*CWCLFLF')
    assert fxn.translate_nt('naaaaa') == 'xK'
    with pytest.raises(OSError):
        fxn.translate_nt('123')

    assert (fxn.rev_translate(''.join(aa).replace('*', ''), codons) ==
            'GCCTGCGACGAGTTCGGCCACATCAAGCTGATGAACCCCCAGAGGAGCACCGTGTGGTAC')


def test_misc_str_handling():
    assert fxn.strip_functionality('()[]test') == 'test'
    assert fxn.tidy_n_term('aaacccggg') == ('aaacccggg', 'KPG')
    assert fxn.tidy_n_term('aaacccgggtt') == ('aaacccggg', 'KPG')
    assert fxn.find_stop('0123456789') == 10
    assert fxn.find_stop('01234*6789') == 5
    assert fxn.dna_check('acgt')
    assert not fxn.dna_check('dna')
    assert fxn.check_suffix_prefix('thisdoesnt', 'overlap') == ''
    assert fxn.check_suffix_prefix('thisdoes', 'doesthough') == 'does'


def test_linker_handling():
    assert fxn.get_linker_dict()
    linkers = fxn.get_linker_dict()
    assert linkers['P2A'] == 'GGCAGCGGCGCCACCAACTTCAGCCTGCTGAAGCAGGCCGGCGACGTGGAGGAGAACCCCGGCCCC'
    linker_seq = fxn.get_linker_seq('P2A', linkers)
    assert linker_seq == 'GGCAGCGGCGCCACCAACTTCAGCCTGCTGAAGCAGGCCGGCGACGTGGAGGAGAACCCCGGCCCC'


def test_junction_determination():
    assert fxn.find_v_overlap('abcdefghijklmno', 'jklmnopqrst') == ('abcdefghi', 'jklmno')
    assert fxn.find_j_overlap('abcdefghijklmno', 'jklmnopqrst') == 'pqrst'

    test_j_motifs = fxn.get_j_motifs('DOG')
    assert isinstance(test_j_motifs[0], dict)
    assert isinstance(test_j_motifs[1], list)
    assert list(set([x[3] for x in test_j_motifs[1]])) == ['J']
    assert coll.Counter(test_j_motifs[0].values()).most_common(1)[0][0] == 'F'

    test_c_motifs = fxn.get_c_motifs('DOG')
    assert isinstance(test_c_motifs, dict)
    assert list(test_c_motifs.keys()) == ['start', 'stop', 'exons']
    assert test_c_motifs['start']['TRAC*01'] == 'VQDPDPSVYQ'
    assert test_c_motifs['start']['TRBC1*01'] == 'DLQKVTPPTV'
    assert len([x for x in test_c_motifs['stop'].values() if '*' in x]) == len(test_c_motifs['stop'])
    assert test_c_motifs['exons']['TRAC*01'] == 'EX1+EX2+EX3+EX4UTR'
    assert test_c_motifs['exons']['TRBC1*01'] == 'EX1+EX2+EX3+EX4'


def run_stitchr_cli(cmd):
    stdout = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
    return stdout.decode().rstrip()


def check_stitching():
    test1 = "stitchr -s dog -v TRAV9-6*01 -j TRAJ42*01 -cdr3 CALRDRGYYGGSQGRLIF -c TRAC*01 -m NT -sw"
    test2 = "stitchr -s dog -v TRBV20*01 -j TRBJ2-6*01 -cdr3 CGYLQGARYEQYF -c TRBC2*01 -m NT -sw"
    test3 = "stitchr -s dog -v TRGV2-1*01 -j TRGJ3-2*01 -cdr3 CAAWEALRHGWYNKVL -c TRGC3*01_A -m NT -sw"
    assert run_stitchr_cli(test1) == ("ATGAAGTGTCCTCCAGGCTTTGTGATTGCCCTATTCTTCATGGTTGGACAAACCCATGGAGACTCAGTGAACCAGACGGA"
                                      "AGGCCAAGTGACCCTCTCAGAAGAGGCCTTCCTGACTATGAACTGTACCTACTCAGCAACAGGATCCCTCATTCTATTCT"
                                      "GGTATGTCCAGTATCTGGGAGAAGGGCCACAGCTCCTCCTGAAAGCATTAAGAGACAAGGAAAAAGGAAGCAGCAAAGGT"
                                      "TTTGAAGCCACCTTGGACAAAACATCCAGATCCTTCCACTTGAAGAAAGGCTCGGTGCAACTGTCAGACTCAGCTGTGTA"
                                      "CTACTGTGCCCTGAGAGACAGGGGCTACTATGGAGGCAGCCAAGGACGACTCATCTTTGGAAAAGGCACTCAACTCTCCG"
                                      "TTAAACCAAATGTCCAGGACCCTGACCCCTCTGTGTACCAGATGAGAAGCCCTAAAACCGGTAGGACTGTCTGCCTGTTC"
                                      "ACCGATTTTGATTCTCAAGCTGATTTCCAAAAACCGGAGGGGGCCACAGTGATCAGCTTGAACAGTACTGTGCTGGACAT"
                                      "GAAGGTTATGGATTCCAAGAGCAATGGGGCCCTGACCTGGAACAGCGAGACTAATATTGAATGCAAGGAAAACTTCCAGC"
                                      "AGCCCTTCTACCCCAGTTCAAACTATTCCTGTGATGCCAAGTTGATAGAGAAAAGCTTTGAAACAGATATGGACCTAAAC"
                                      "TTCTACAACCTGCTAGTGATTGGGCTACGCATCCTCCTCCTGAAAGTGATCGGATTTAACCTGTTCATGACGCTGCGGCT"
                                      "GTGGTCCTGC")
    assert run_stitchr_cli(test2) == ("ATGCTGATGCTTCTGCTGCTCCTGGGGCCCAGCTCTGGACTCGGTGCCCTCGTCTTCCAGGCGCCCAGCACAATGATCTG"
                                      "TAAGAGCGGAGCCACCGTGCAGATCCAGTGTCAAACAGTGGACCTTCAAGCCACAACCGTGTTTTGGTATCGCCAGCTCC"
                                      "CGAAGCAGGGCCTTACCCTTATGGTGACCTCTAACGTGGGCAACAGTGCTACACACGAGCAGGGGTTCCCTGCAGCCAAG"
                                      "TTCCCTGTTAACCACCCAAACCTCACGTTTTCCTCCCTGATGGTGACGAGTTCAGGTCCTGGAGACAGCGGCCTCTACTT"
                                      "CTGTGGTTACCTGCAGGGCGCCCGCTACGAGCAGTATTTCGGCGCCGGCACCAGGCTCACGGTCCTCGAGGATCTGCAGA"
                                      "AGGTCACCCCTCCCACGGTCACAGTGTTTGAACCATCAGAAGCAGAGATCCCGCGGACCCAGAAGGCCACACTCGTGTGC"
                                      "CTGGCCACGGGCTTCTACCCCGACCACGTGGAGCTGAGCTGGTGGGTGAACGGGAAGGAGGTCACGAGTGGGTTCAGCAC"
                                      "CGACCCGCAGCCCTACAAGGAGAGGCCCAGCGAGAATGACTCCAGCTACTGTCTGAGCAGCCGGCTGAGGGTCTCTGCCT"
                                      "CCTTCTGGCACAACCCGCGCAACCACTTCCGCTGCCAAGTCCAGTTCTATGGGCTCGGGGACGACGATGAGTGGAAATAC"
                                      "GATAGAGTCAAACCCATCACCCAGAACATCAGTGCTGAGGCCTGGGGCAGAGCAGACTGTGGCTTCACCTCGGTGTCCTA"
                                      "CCATCAGGGCGTCCTGTCTGCCACCATCCTCTATGAGATCCTGCTGGGCAAGGCCACGCTGTATGCTGTGCTGGTCAGCA"
                                      "TCCTGGTGCTGATGGCCAAGGTCAAGAGAAAAGGTTCC")
    assert run_stitchr_cli(test3) == ("ATGCTGAGTTCCCTTGCTGTCCTGTGTGTCCTCCTGTTTCCCGGTGGTGCAGCGGGGCTCAAGCTGGAGCAGACTCCTGT"
                                      "GGTCGTGGGGCGCTTGGGCACCTTGGCCACCCTGCCGTGCACAGTGGATACCTCGGTGAGCTACATCCACTGGTACTTCC"
                                      "ACAAGGAGGATACAGCCCCCAAGAGGCTCCTCTACCTAGACATGTCCAGGTCGTATGTGCAGAGGGATATATTCGTGAAA"
                                      "GCAGACAAAGTCAACGCCAAGAAAGGCAGGAACAGTTACAGCTGTAACCTGTTGTTGCAGAAACTGGAGAAGAATGATGA"
                                      "GGGCGTGTACTACTGCGCTGCCTGGGAGGCCCTGAGGCACGGCTGGTACAACAAAGTGCTTGGTCCTGGCACAATTCTCA"
                                      "GGGTTACAGATAAAAGTGCCAATGAAGACACTTCCCCCAAGCCCACAGTTTTTCTTCCTTCGATTTCTGAAATAAAAATC"
                                      "CATAAGGCTGGAACATATCTTTGTCTTCTCGAGGATTTTTTTCCTGAAATTATCAAGGTAGATTGGAAAGAAAAGAATGA"
                                      "CCAAACAGTTCTGCAATCCCAGCAGGGAAACACTATGAAGACTAAGGACACATACATGAAATTCAGCTGGCTGACCGTGA"
                                      "CTGGAGCGTCTATGGATAAAGAACACAAATGTATTGTCAACCACGAGAGTGATAGAGGAGGGATTAATCAAGAGATTCTT"
                                      "TTTCCTTCGATAAATGAAGATTTGGTAGCGAGGATGGAATCTGATGTTGACTCCGAAGGGCATTCAGAGGCAAAGCAAAA"
                                      "CACAGAGGTGGTAACTGTGAGCCCTCTAAGTTCACGATCTTTCCCACCTGGCGCAAGCACTGTGGACCCCCTGCAGCTGC"
                                      "AGCTCATGAACACCTCTGCCTACTATACCTACCTCCTCCTCCTCCTGAAGAGTCTGACGTACTCCATCATCATCACCATC"
                                      "TATCTGCTTGGGAGATCAGTGCTCAATGGTAATGGGAAGAGTTCC")


# TODO STILL
    # def get_additional_genes(imgt_data, imgt_functionality):
    # def tidy_c_term(c_term_nt, skip, c_region_motifs, c_gene):
    # def determine_v_interface(cdr3aa, n_term_nuc, n_term_amino):
    # def find_cdr3_c_term(cdr3_chunk, j_seq, strict):
    # def determine_j_interface(cdr3_cterm_aa, c_term_nuc, c_term_amino, gl_nt_j_len, j_warning_threshold):
    # def tweak_thimble_input(stitch_dict): # Todo move to a thimble test file?

    # Also need some overall stitching tests! (NT / seamless)